import json
import logging
import os
import random
import string
import subprocess
import urllib
import uuid
import ssl
from flask import Response, send_from_directory, Blueprint, request, abort
from barix.web.auth import login_required
from barix.web.system_info import getStorageSize
from barix.web.uci import getValueOfUci, setUciConfigs

from restart_services import createListOfServicesToRestart

bp = Blueprint('audio_files_API', __name__)
mainUrl = "http://127.0.0.1:8765"

log = logging.getLogger('flask-backend')

def buildResponse(headers, status, payload=None):
    response_headers = {}

    for (key, value) in headers.items():
        if key.lower() in ["content-length", "connection"]:
            continue
        response_headers[key] = value
    if payload is None:
        return Response(status=status, headers=response_headers)
    else:
        return Response(response=payload, status=status, headers=response_headers)


@bp.route('/audio-files/pregong', methods=['POST'])
@login_required
def new_pregong_file():
    filename = ""
    wav_filename = ""
    try:
        filesUploaded = request.files

        if 'file' not in filesUploaded:
            log.error("Invalid file uploaded")
            return Response("Invalid file uploaded", status=400)

        uploadedFile = filesUploaded['file']
        filename = uploadedFile.filename
        try:
            if filename == '':
                log.error("Invalid filename")
                # return Response("Invalid filename", status=400)
                raise Exception("Invalid filename")

            tmp_path = os.path.join('/tmp', filename)

            uploadedFile.save(tmp_path)

            # check if file size is larger than available free space in storage (because of 5% root dedicated space, file is stored even if available space shows as 0, since backed is being run as root)
            if os.stat(tmp_path).st_size > getStorageSize()["available"]:
                # return Response('No space left on device', status=500)
                raise Exception("No space left")

            # move to final destination
            final_filepath = os.path.join("/mnt/data/pregong", filename)
            subprocess.run(['mv', tmp_path, final_filepath])

        except Exception as e:
            if str(e) == "Invalid filename":
                return Response("Invalid filename", status=400)
            elif str(e) == "No space left":
                return Response('No space left on device', status=500)
            else:
                return Response(status=500)

        # convert audio file to wav
        sample_rate = getValueOfUci('icpgw', 'barp', 'sample_rate')
        format = "pcm_mulaw"

        wav_filename = f"{os.path.splitext(filename)[0]}.wav"

        command = [
            "ffmpeg",
            "-y",  # to overwrite the output file if exists.
            "-i",
            final_filepath,
            "-acodec",
            f"{format}",
            "-ar",
            f"{sample_rate}",
            "-ac",
            "1",
            os.path.join("/mnt/data/pregong/processed", wav_filename),
        ]
        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        try:
            outs, errs = process.communicate(timeout=15)  # Longer timeout so larger files can be processed
            # log.debug(outs)
            # log.debug(errs)
        except subprocess.TimeoutExpired:
            process.kill()
            outs, errs = process.communicate()
            log.error(
                f"Timed out, killing processing of {filename} : errs :{errs}"
            )
            # log.warning("To avoid this, try increasing the timeout so larger files can be processed")
        except Exception as e:
            log.error(f"Error converting pregong file to wav: {e}")


        # remove previous file (if any)
        for file in os.listdir("/mnt/data/pregong"):
            complete_path = os.path.join("/mnt/data/pregong", file)
            if os.path.isfile(complete_path) and file != filename:
                log.info(f"Deleting previous pregong file {file}")
                os.remove(complete_path)
        for file in os.listdir("/mnt/data/pregong/processed"):
            complete_path = os.path.join("/mnt/data/pregong/processed", file)
            if os.path.isfile(complete_path) and file != wav_filename:
                log.info(f"Deleting previous processed pregong file {file}")
                os.remove(complete_path)

        return Response(status=200)
    except Exception as e:
        log.error(e)
        log.info("New device message uploaded failed! Removing physical audio files")
        raw_file_path = os.path.join("/mnt/data/pregong", filename)
        processed_file_path = os.path.join("/mnt/data/pregong/processed", wav_filename)
        if os.path.isfile(raw_file_path):
            os.remove(raw_file_path)
        if os.path.isfile(processed_file_path):
            os.remove(processed_file_path)
        return Response(status=500)


@bp.route('/audio-files/pregong/info', methods=['GET'])
@login_required
def get_pregong_file_info():
    try:
        response_json = {"filename": ""}
        for file in os.listdir("/mnt/data/pregong"):
            if os.path.isfile(os.path.join("/mnt/data/pregong", file)):
                response_json["filename"] = file
        return Response(json.dumps(response_json), status=200, headers={"Content-Type": "application/json"})
    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/audio-files/pregong', methods=['DELETE'])
@login_required
def delete_pregong_file():
    try:
        for file in os.listdir("/mnt/data/pregong"):
            complete_path = os.path.join("/mnt/data/pregong", file)
            if os.path.isfile(complete_path):
                log.info(f"Deleting previous pregong file {file}")
                os.remove(complete_path)
        for file in os.listdir("/mnt/data/pregong/processed"):
            complete_path = os.path.join("/mnt/data/pregong/processed", file)
            if os.path.isfile(complete_path):
                log.info(f"Deleting previous processed pregong file {file}")
                os.remove(complete_path)

        # force enabled to be a false when deleting pre-gong file
        setUciConfigs( {"icpgw.pregong.enabled": False}, commit=True, restartServices=True,
                                       selectServicesToRestartCallback=createListOfServicesToRestart)

        return Response(status=200)
    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/audio-files', methods=['GET'])
@login_required
def get_all_audio_files():
    try:
        # get users from main backend
        context = ssl.create_default_context()
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        redirection_req = urllib.request.Request(url=mainUrl + "/audio_files", method="GET")
        res = urllib.request.urlopen(redirection_req, context=context)
        data = res.read()
        enc = res.headers.get_content_charset('utf-8')
        payload = data.decode(enc)
        json_payload = json.loads(payload)

        # transform json into list of users
        audioFilesList = []
        for key, value in json_payload.items():
            audioFileDict = {"id": key}
            for key2, value2 in value.items():
                audioFileDict[key2] = value2
            audioFilesList.append(audioFileDict)

        # send list of users in response
        response_json = {"audio-files": audioFilesList}
        return Response(json.dumps(response_json), status=200, headers={"Content-Type": "application/json"})
    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/audio-files', methods=['POST'])
@login_required
def new_audio_file():
    final_filename = ""
    try:
        filesUploaded = request.files

        if 'audio_file' not in filesUploaded:
            log.error("Invalid file uploaded")
            return Response("Invalid file uploaded", status=400)

        uploadedFile = filesUploaded['audio_file']
        try:
            final_filename = handleFileUploaded(uploadedFile)
        except Exception as e:
            if str(e) == "Invalid filename":
                return Response("Invalid filename", status=400)
            elif str(e) == "No space left":
                return Response('No space left on device', status=500)
            else:
                return Response(status=500)

        request_data = json.loads(request.form['data'])

        payload = {
            "btn_name": request_data["name"],
            "file_name": final_filename,
            "modbus_idx": request_data["modbus_idx"]
        }

        # send request to main app
        context = ssl.create_default_context()
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        redirection_req = urllib.request.Request(url=mainUrl + "/audio_files", data=json.dumps(payload).encode('utf-8'), method="POST")
        res = urllib.request.urlopen(redirection_req, context=context)

        res_body = res.read()
        enc = res.headers.get_content_charset('utf-8')
        new_audio_file_id = json.loads(res_body.decode(enc))["id"]

        res_code = res.getcode()

        # update users with audio file if audio_file upload was successful
        if res_code == 201 and len(request_data["users"]) > 0:
            # get users
            redirection_req = urllib.request.Request(url=mainUrl + "/users", method="GET")
            res2 = urllib.request.urlopen(redirection_req, context=context)
            data = res2.read()
            enc = res2.headers.get_content_charset('utf-8')
            payload = data.decode(enc)
            users_dict = json.loads(payload)

            # for each user
            for user_id in users_dict:
                # if user id is in list from request
                if user_id in request_data["users"]:
                    # add it
                    user_data = users_dict[user_id]
                    if "audio_files" in user_data.keys():
                        user_data["audio_files"].append(new_audio_file_id)
                    else:
                        user_data["audio_files"] = [new_audio_file_id]

                    # update user
                    user_data["id"] = user_id

                    redirection_req = urllib.request.Request(url=mainUrl + "/users",
                                                             data=json.dumps(user_data).encode('utf-8'), method="UPDATE")
                    urllib.request.urlopen(redirection_req, context=context)

        return buildResponse(res.headers, res_code, res_body)
        # return Response(status=200)
    except Exception as e:
        log.error(e)
        log.info("New device message uploaded failed! Removing physical audio files (if any)")
        raw_file_path = os.path.join("/mnt/data/prerecorded_messages", final_filename)
        processed_file_path = os.path.join("/mnt/data/prerecorded_messages/processed", (final_filename + ".wav"))
        if os.path.isfile(raw_file_path):
            os.remove(raw_file_path)
        if os.path.isfile(processed_file_path):
            os.remove(processed_file_path)
        return Response(status=500)


@bp.route('/audio-files', methods=['PUT'])
@login_required
def edit_audio_file():
    try:
        request_data = json.loads(request.form['data'])

        payload = {
            "btn_id": request_data["id"],
            "btn_name": request_data["name"],
            "file_name": request_data["filename"],
            "modbus_idx": request_data["modbus_idx"]
        }

        filesUploaded = request.files

        if 'audio_file' in filesUploaded:
            log.info("Audio file uploaded: updating audio file associated with ID {}".format(request_data["id"]))
            # log.error("Invalid file uploaded")
            # return Response("Invalid file uploaded", status=400)

            uploadedFile = filesUploaded['audio_file']

            try:
                final_filename = handleFileUploaded(uploadedFile)
                payload["file_name"] = final_filename
            except Exception as e:
                if str(e) == "Invalid filename":
                    return Response("Invalid filename", status=400)
                elif str(e) == "No space left":
                    return Response('No space left on device', status=500)
                else:
                    return Response(status=500)

        # send request to main app
        context = ssl.create_default_context()
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        redirection_req = urllib.request.Request(url=mainUrl + "/audio_files", data=json.dumps(payload).encode('utf-8'),
                                                 method="UPDATE")
        res = urllib.request.urlopen(redirection_req, context=context)

        res_code = res.getcode()

        if res_code == 200:
            # update users with audio file
            # get users
            redirection_req = urllib.request.Request(url=mainUrl + "/users", method="GET")
            res2 = urllib.request.urlopen(redirection_req, context=context)
            data = res2.read()
            enc = res2.headers.get_content_charset('utf-8')
            payload = data.decode(enc)
            users_dict = json.loads(payload)

            # for each user
            changed = False
            for user_id in users_dict:
                # if user id is in list from request
                if user_id in request_data["users"]:
                    # add it
                    user_data = users_dict[user_id]
                    if "audio_files" in user_data.keys():
                        user_data["audio_files"].append(request_data["id"])
                    else:
                        user_data["audio_files"] = [request_data["id"]]
                    changed = True
                # else
                else:
                    # if it exists
                    if "audio_files" in users_dict[user_id].keys() and request_data["id"] in users_dict[user_id]["audio_files"]:
                        user_data = users_dict[user_id]
                        # remove it
                        user_data["audio_files"] = [x for x in user_data["audio_files"] if x != request_data["id"]]
                        changed = True
                if changed:
                    # update user
                    user_data["id"] = user_id
                    redirection_req = urllib.request.Request(url=mainUrl + "/users",
                                                             data=json.dumps(user_data).encode('utf-8'),
                                                             method="UPDATE")
                    urllib.request.urlopen(redirection_req, context=context)

        return buildResponse(res.headers, res_code, res.read())
    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/audio-files', methods=['DELETE'])
@login_required
def delete_audio_file():
    try:
        # get users list of ids to delete from request body
        audioFilesList = request.json
        payload = {"audio_file_ids": audioFilesList}

        # send request to main app
        context = ssl.create_default_context()
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        redirection_req = urllib.request.Request(url=mainUrl + "/audio_files", data=json.dumps(payload).encode('utf-8'), method="DELETE")
        res = urllib.request.urlopen(redirection_req, context=context)

        return buildResponse(res.headers, res.getcode(), res.read())
    except Exception as e:
        log.error(e)
        return Response(status=500)


def handleFileUploaded(fileUploaded):
    # if user does not select file, browser also
    # submit an empty part without filename
    if fileUploaded.filename == '':
        log.error("Invalid filename")
        # return Response("Invalid filename", status=400)
        raise Exception("Invalid filename")

    # generate random file name
    tmp_filename = str(uuid.uuid4())
    tmp_path = os.path.join('/tmp', tmp_filename)

    fileUploaded.save(tmp_path)

    # check if file size is larger than available free space in storage (because of 5% root dedicated space, file is stored even if available space shows as 0, since backed is being run as root)
    if os.stat(tmp_path).st_size > getStorageSize()["available"]:
        # return Response('No space left on device', status=500)
        raise Exception("No space left")

    # move to final destination
    final_filepath = os.path.join("/mnt/data/prerecorded_messages", tmp_filename)
    subprocess.run(['mv', tmp_path, final_filepath])

    return tmp_filename


