# standard library imports
import logging
import os
import sys
import threading
from datetime import timedelta

from flask import Flask, render_template
from pathlib import Path
# related third party imports

# local application/library specific imports
import logger

from barix.web.restart_services import restartServicesFromFile
from barix.web.constants import SSL_CERTIFICATES_DIR
from barix.web.uci import getValueOfUci
from barix.web.certificates import generateSSLCerts, certificateIsValid, keyIsValid

UPLOAD_FILES_DIR = "/mnt/data/uploads"
CONFIG_DIR = "/mnt/data/config"
PREGONG_DIR = "/mnt/data/pregong"
PROCESSED_PREGONG_DIR = "/mnt/data/pregong/processed"
MAX_FILE_SIZE_UPLOAD = 256 * 1024 * 1024 #256Mb

def include_routes_to_frontend():
    try:
        from static.frontendRoutes import frontendRoutes
        flask_app.register_blueprint(frontendRoutes)
        #log.debug("Imported frontend routes successfully")
    except Exception as e:
        log.debug("It was not possible to import frontend routes: {}".format(e))

def include_standard_routes():
    import settings_API
    flask_app.register_blueprint(settings_API.bp)
    import auth_API
    flask_app.register_blueprint(auth_API.bp)
    import logs_API
    flask_app.register_blueprint(logs_API.bp)
    import reboot_API
    flask_app.register_blueprint(reboot_API.bp)
    import reset_to_defaults_API as defaults_API
    flask_app.register_blueprint(defaults_API.bp)
    import system_info_API
    flask_app.register_blueprint(system_info_API.bp)
    import uci_API
    flask_app.register_blueprint(uci_API.bp)
    import fw_update_API
    flask_app.register_blueprint(fw_update_API.bp)
    import ca_certificates_API
    flask_app.register_blueprint(ca_certificates_API.bp)
    import paging_API
    flask_app.register_blueprint(paging_API.bp, url_prefix='/api/rest/v1')
    import users_API
    flask_app.register_blueprint(users_API.bp, url_prefix='/api/rest/v1')
    import groups_API
    flask_app.register_blueprint(groups_API.bp, url_prefix='/api/rest/v1')
    import audio_files_API
    flask_app.register_blueprint(audio_files_API.bp, url_prefix='/api/rest/v1')

from io import BytesIO
from tempfile import TemporaryFile, SpooledTemporaryFile
from flask.wrappers import Request

class MemoryRequest(Request):
    def _get_file_stream(self, total_content_length, content_type, filename, content_length):
        max_size = 1024 * 500
        if SpooledTemporaryFile is not None:
            return SpooledTemporaryFile(max_size=max_size, mode="rb+", dir=UPLOAD_FILES_DIR)
        elif total_content_length is None or total_content_length > max_size:
            return TemporaryFile(mode="rb+", dir=UPLOAD_FILES_DIR)
        return BytesIO()

"""
Init logger
"""
logger.configLogger('flask-backend', "/var/log/webui-backend.log", 2*1024*1024, 2)
log = logging.getLogger('flask-backend')
log.info("Starting flask backend v2.0...")

"""
Check if has services to restart
"""
restartServicesFromFile()

"""
Create non existing directories
"""
dir_list = [UPLOAD_FILES_DIR, CONFIG_DIR, PREGONG_DIR, PROCESSED_PREGONG_DIR]
for dir in dir_list:
    p = Path(dir)
    if not p.exists():
        try:
            p.mkdir()
        except Exception as e:
            log.error("Error creating directory {}: {}".format(dir, e))

# p = Path(UPLOAD_FILES_DIR)
# if not p.exists():
#     try:
#         p.mkdir()
#         #log.debug("Directory {} successfully created".format(UPLOAD_FILES_DIR))
#     except Exception as e:
#         log.error("Error creating directory {}: {}".format(UPLOAD_FILES_DIR, e))
#
# p = Path(CONFIG_DIR)
# if not p.exists():
#     try:
#         p.mkdir()
#     except Exception as e:
#         log.error("Error creating directory {}: {}".format(CONFIG_DIR, e))


"""
Init Flask application
"""
# static_url_path='' removes any preceding path from the URL (i.e. the default /static).
# static_folder='web/static' to serve any files found in the folder web/static as static files.
# template_folder='web/templates' similarly, this changes the templates folder.
flask_app = Flask(__name__, static_url_path='', static_folder='static', template_folder='static')
# secret key used to sign session cookies for protection against cookie data tampering.
# TODO: get secret key from anywhere else, like env variable, for security (?)
flask_app.secret_key = b'\xb3\xa482C\x98\x81\xa9\x8e5\xe9\x7f\xe2\xf9y/'
# UPLOAD_FOLDER is where uploaded files are stored
flask_app.config['UPLOAD_FOLDER'] = UPLOAD_FILES_DIR
# maximum size allowed for the files to upload
flask_app.config['MAX_CONTENT_LENGTH'] = MAX_FILE_SIZE_UPLOAD
# override default werkzeug formparser to allow to update files to the UPLOAD_FOLDER instead of default
flask_app.request_class = MemoryRequest

"""
Include routes
"""
include_routes_to_frontend()
include_standard_routes()

"""
Implement session login timeout is login is required
"""
if getValueOfUci('httpd', 'webserver', 'password_set') == "true":
    session_timeout = getValueOfUci('httpd', 'webserver', 'session_timeout')
    if session_timeout == "":
        session_timeout = "30"
    log.info("Using session timeout value: {}".format(session_timeout))
    flask_app.config['PERMANENT_SESSION_LIFETIME'] = timedelta(minutes=int(session_timeout))
else:
    log.info("Login is not required => no session timeout.")

"""
Start flask
"""
priv_key = os.path.join(SSL_CERTIFICATES_DIR, 'barix.pem')
cert = os.path.join(SSL_CERTIFICATES_DIR, 'barix.pem')
notRunning = True
while(notRunning):
    try:
        if os.path.isfile(priv_key) and os.path.isfile(cert) and certificateIsValid(cert) and keyIsValid(priv_key):
            flask_app.run(host='0.0.0.0', port=443, threaded=True, ssl_context=(cert, priv_key))
            log.debug("Flask app started!")
            notRunning = False
        else:
            log.warning("File {} or {} don't exist or are not valid!".format(priv_key, cert))
            log.warning("Generating new SSL certificates...")
            generateSSLCerts()
    except Exception as e:
        log.error(e)