from utils import read_json, get_logger
import json
from typing import Dict, List, Union, Tuple, NewType


logger = get_logger("Groups")


class TYPES:
    client_id = NewType("client_id", int)
    group_id = NewType("group_id", str)
    group_name = NewType("group_name", str)
    group = Dict[str, Union[List[client_id], group_name]]
    groups = Dict[group_id, group]


class Groups:
    TOTAL_GROUPS = set([str(i) for i in range(1, 129)])

    def __init__(self):
        self._groups = read_json("groups_backup.json")
        self.occupied_group_id = set(self._groups.keys())

    @property
    def value(self) -> TYPES.groups:
        return self._groups

    @property
    def new_id(self) -> Union[None, TYPES.group_id]:
        """Gets an unoccupied group id.

        Returns:
            Union[str, None]: returns id or None depending upon availability.
        """
        available_ids = Groups.TOTAL_GROUPS - self.occupied_group_id
        available_ids_int = set(map(int, available_ids))  # Otherwise the min value is not correct
        min_id = TYPES.group_id(str(min(available_ids_int)))
        # logger.warning("Available group IDs: {}".format(available_ids_int))
        logger.warning("Min group ID as TYPES.group_id: {}".format(min_id))
        if len(available_ids) != 0:
            # return min(available_ids)
            return min_id
        else:
            return None

    @property
    def bytes(self) -> bytes:
        return json.dumps(self._groups).encode()

    def get_clients(self, group_ids: List[TYPES.group_id]) -> List[TYPES.client_id]:
        client_ids = set([])
        if group_ids is None:
            logger.error("################### get_clients returned None ")
            return []
        for group_id in group_ids:
            client_ids.update(self._groups[str(group_id)]["clients"])
        return list(client_ids)

    # def validate_clients(self, clients: List[TYPES.client_id]) -> Tuple[bool, str]:
    def validate_clients(self, clients: Union[List[TYPES.client_id], TYPES.group_name]) -> Tuple[bool, str]:
        # type check:
        if not isinstance(clients, list):
            return False, "Not a client list"
        elif len(clients) == 0:
            return False, "Empty client list"
        else:
            return True, "OK"

    def add(self, group: TYPES.group) -> Tuple[bool, Union[dict, str]]:
        group_name = group["name"]
        clients = group["clients"]
        valid_clients, msg = self.validate_clients(clients)
        new_id = self.new_id
        if new_id is not None and valid_clients:
            self._groups[new_id] = {"name": group_name, "clients": clients}
            self.occupied_group_id.add(new_id)
            self.save()
            return True, {"id": new_id}
        else:
            return False, "Improper Request"

    def update(self, group: TYPES.group) -> Tuple[Union[int, bool], str]:
        group_name = group["name"]
        clients = group["clients"]
        group_id = str(group["id"])
        valid_clients, msg = self.validate_clients(clients)
        if group_id in self._groups:
            if valid_clients:
                self._groups[group_id] = {"name": group_name, "clients": clients}
                self.save()
                return True, "OK"
            else:
                return False, "Clients are not valid"
        else:
            return -1, "group doesn't exist!"

    def remove(self, group):
        group_ids = group["group_ids"]
        for group_id in group_ids:
            if str(group_id) in self._groups:
                del self._groups[str(group_id)]
                self.refresh_occupied_ids()
                logger.warning(f"Remove : Success {group_id}")
                # self.save()
            else:
                logger.warning(
                    f"Group deletion unsucessfull, {group_id} doesn't exist"
                )
        self.save()
        return True, "OK"

    def save(self):
        with open("groups_backup.json", "w") as f:
            json.dump(self._groups, f)

    def refresh_occupied_ids(self):
        self.occupied_group_id = set(self._groups.keys())
