import audioop
import datetime
import json
import re
import socketserver
import threading
import time
import timeit

from http.server import BaseHTTPRequestHandler, HTTPServer
from http import HTTPStatus

from secrets import token_urlsafe
from threading import Thread, enumerate
from typing import Tuple, Union

import shared_variables
from utils import get_logger, check_wav_file_format, store_alsa_settings

from paging_functions import local_playback_stop, encode_send_audio, extract_token_and_audio
from paging_functions import start_paging_message, stop_paging_message
from paging_functions import start_playing_file, stop_playing_file

# Anton imports
import os
import sys
import constants
import traceback
from multiprocessing import Process
from queue import Queue

logger = get_logger("HTTP_Server_backend")

class PagingHttpRequestHandler(BaseHTTPRequestHandler):

    def _set_response(self, code: int = 200, message: str = ""):
        logger.debug(f"_set_response {code} - {message}")
        self.send_response(code, message)
        if message != "":
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps({"response": message}).encode())
        else:
            self.send_header("Content-type", "text/html")
            self.end_headers()
    
    def _set_response_encoded(self, code: int = 200, data=None):
        logger.debug(f"_set_response {code} - {data}")
        self.send_response(code, data)    
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.wfile.write(data)

    def _check_all_permissions(self, message):
        allow_paging = True
        uuid = None

        try:
            uuid = message['uuid']
        except Exception as e:
            logger.error(f"Following occurred: {e}")
            #self._set_response(code=HTTPStatus.UNAUTHORIZED.value)
        
        if shared_variables.modbus_running == True:
            error_msg = "Modbus running"
            logger.warning(error_msg)
            self._set_response(code=HTTPStatus.CONFLICT.value, message=error_msg)
            return False

        if not shared_variables.paging_controller._check_user(message["token"]):
            # logger.info(f"User matched: {shared_variables.active_user}")
            error_msg = f"Another user is paging: {shared_variables.active_user}"
            new_user = shared_variables.users.get_user_from_token(message["token"])
            logger.info(f"Received request from {new_user}")
            logger.warning(error_msg)
            front_end_reply = "Another user is paging!"
            self._set_response(code=HTTPStatus.CONFLICT.value, message=front_end_reply)
            return False

        if uuid is not None:
            allow_paging = shared_variables.paging_controller.is_operation_allowed(uuid, "Paging")

        # logger.info("Message type: %s", message["type"])

        if not allow_paging:
            logger.warning(f"Same user trying to page from a different UUID not allowed!"
                            f"\nRequest -> {message},\nCurrent UUID -> {shared_variables.active_user_uuid}")
            error_msg = "Same user trying to page from a different device not allowed!"
            self._set_response(code=HTTPStatus.CONFLICT.value, message=error_msg)
            return False
        
        if not shared_variables.paging_controller._device_idle() and message["type"] != 8:
            error_msg = "Operation must be first stopped from all users and devices before proceeding."
            logger.warning(error_msg)
            self._set_response(code=HTTPStatus.CONFLICT.value, message=error_msg)
            return False

        return True
    
    def get_file_to_open(self, message):


        file_id = check_wav_file_format(message["file_ID"])

        files = []  # Store all the files from the given directory
        for file in os.listdir(constants.PROCESSED_FILES_PATH):
            files.append(file)


        # Check if the requested file has a physical equivalent in storage
        if file_id not in files:
            logger.warning("The requested file ID: %s does not exist", file_id)
            shared_variables.playfile_state = constants.IDLE  #
            shared_variables.kill_playfile_thread = constants.YES
            shared_variables.audio_files.set_active_audio_file(None)
        else:
            shared_variables.downloading_recorded_message = constants.NO
            shared_variables.first_fast_packets = 0
            shared_variables.kill_playfile_thread = constants.NO

            logger.debug(f"File {file_id} found in existing files")
            if "button_ID" in message:
                shared_variables.audio_files.set_active_audio_button_id(message["button_ID"])
            else:
                shared_variables.audio_files.set_active_audio_file(file_id)
            shared_variables.playfile_state = constants.BUSY

        return os.path.join(constants.PROCESSED_FILES_PATH, file_id)
    
#======================================================================================================================
#
#======================================================================================================================

    def do_GET(self):
        logger.debug("Received a GET request. path = %s", self.path)
        try:
            if self.path == "/clients":
                self._return_active_clients()
            elif self.path == "/groups":
                self._return_groups()
            elif self.path == "/audio_files":
                self._return_audio_files()
            elif self.path == "/users":
                self._return_users()
            else:
                self._set_response(code=404, message="Unknown GET request path")
        except Exception as e:
            logger.error(f"Exception occurred while executing GET {self.path} request")
            logger.error(f"{e}")

    def do_PUT(self):
        logger.debug("Received a PUT request. path = %s", self.path)
        try:
            if self.path == "/paging":
                self._page_audio()
            elif self.path == "/recording":
                self._record_audio()
            elif self.path == "/playfile":
                self._get_playfile_state()
            else:
                self._set_response(code=404, message="Unknown PUT request path")
        except Exception as e:
            logger.warning(f"Exception occurred while executing PUT {self.path} request")
            logger.warning(f"{e}")

    def do_POST(self):
        """
        Post request for starting/stopping paging request.
        """
        logger.debug("Received a POST request. path = %s", self.path)
        try:
            if self.path == "/paging":
                self._start_stop_page_message()
            elif self.path == "/recording":
                self._start_stop_page_message()

            elif self.path == "/playfile":
                self._start_stop_playing_file()
            elif self.path == "/groups":
                self._create_group()
            elif self.path == "/users":
                self._create_user()
            elif self.path == "/audio_files":
                self._create_audio_file()
            elif re.match('\/users\/\d+\/token', self.path):
                self._renew_user_token()
            else:
                self._set_response(code=404, message="Unknown POST request path")
        except Exception as e:
            logger.error(f"Exception occurred while executing POST {self.path} request")
            logger.error(f"{e}")

    def do_DELETE(self):
        try:
            if self.path == "/groups":
                self._delete_group()
            elif self.path == "/users":
                self._delete_user()
            elif self.path == "/audio_files":
                self._delete_audio_file()
            else:
                self._set_response(code=404, message="Unknown DELETE request path")
        except Exception as e:
            logger.error(f"Exception occurred while executing DELETE {self.path} request")
            logger.error(f"{e}")


    def do_UPDATE(self):
        try:
            if self.path == "/groups":
                self._update_group()
            elif self.path == "/audio_files":
                self._update_audio_file()
            elif self.path == "/users":
                self._update_user()
            else:
                self._set_response(code=404, message="Unknown UPDATE request path")
        except Exception as e:
            logger.error(f"Exception occurred while executing UPDATE {self.path} request")
            logger.error(f"{e}")

#======================================================================================================================
# do_GET
#======================================================================================================================

    def _return_active_clients(self):
        self.send_response(code=200)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.wfile.write(
            json.dumps({"active_clients": list(shared_variables.pmaster.active_devices.keys())}).encode()
        )

    def _return_users(self):
        try:
            self._set_response_encoded(code=HTTPStatus.OK.value, data=shared_variables.users.bytes)
        except Exception as e:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.error(message)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=message)

    def _return_groups(self):
        try:
            self._set_response_encoded(code=HTTPStatus.OK.value, data=shared_variables.groups.bytes)
        except Exception as e:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.error(message)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=message)

    def _return_audio_files(self):
        try:
            self._set_response_encoded(code=HTTPStatus.OK.value, data=shared_variables.audio_files.bytes)
        except Exception as e:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.error(message)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=message)

#======================================================================================================================
# do_PUT
#======================================================================================================================

    def _page_audio(self):
        shared_variables.downloading_recorded_message = constants.NO
        content_length = int(
            self.headers["Content-Length"]
        )  # <--- Gets the size of data
        message = self.rfile.read(content_length)
        token, audio = extract_token_and_audio(message)

        logger.debug("_page_audio")

        shared_variables.watchDog_live_paging.reset()

        if shared_variables.modbus_running == True:
            error_msg = "Modbus running"
            logger.warning(error_msg)
            self._set_response(code=HTTPStatus.CONFLICT.value, message=error_msg)
            return


        # This blocks all web requests
        #while shared_variables.pregong_playing == True :
        #    time.sleep(0.001)
        if shared_variables.pregong_playing == True :
            self._set_response(code=HTTPStatus.NO_CONTENT.value)
            return

        try:
            if token == shared_variables.users.secret(shared_variables.active_user):
                
                # send to groups
                if shared_variables.pmaster.state == 130:
                    shared_variables.paging_message_q.put(audio)  # Add data in the que
                    shared_variables.paging_message_q_buff_msg += 1
                    # logger.info(f"Added to queue {len(audio)}, msgs: {shared_variables.paging_message_q_buff_msg}")

                    # Start the queue reading thread as soon as there is data in the queue. Always 16Kb in the buff.
                    # This way prevent problems with very short button clicks.
                    if shared_variables.paging_message_q_buff_msg == 1:
                        shared_variables.paging_message_q_thread.start()  # This thread is only reading from the queue


                # local playback
                try:
                    shared_variables.local_playback_msg_queue.put(audio)
                except Exception as e:
                    logger.warning(f"Couldn't add to local playback queue: {e}")
                
                if shared_variables.paging_message_q_buff_msg == 1:
                    shared_variables.local_playback_queue_process.start()
                    # Start reading from the que once there is data

                #pregong_found, file_name = shared_variables.check_pregong()
                #if not pregong_found and shared_variables.paging_message_q_buff_msg > 0 and shared_variables.local_playback_queue_process.is_alive()==False:
                #    shared_variables.local_playback_queue_process.start()

                #if shared_variables.pregong_playing == False and shared_variables.paging_message_q_buff_msg > 0 and shared_variables.local_playback_queue_process.is_alive()==False:
                #    shared_variables.local_playback_queue_process.start()


                #print(f"        CHECK THREAD {shared_variables.local_playback_queue_process}")
                #print(dir(shared_variables.local_playback_queue_process))
                #print(f"Thread {shared_variables.local_playback_queue_process.is_alive()}" )



                time.sleep(0.001)  # Give chance to other processes to work [10ms]
                self._set_response(code=HTTPStatus.NO_CONTENT.value)
            else:
                logger.warning("User token not found!")

        except KeyError as e:
            local_playback_stop()
            # The request could not be processed because of conflict in the current state of the resource,
            if shared_variables.active_user is None:
                logger.error(f"Key error. The current user is: {e}")
            else:
                logger.error(f"Key error. Message: {e}")
                traceback.print_exc()
            self._set_response(HTTPStatus.INTERNAL_SERVER_ERROR.value)


    def _record_audio(self):
        # local_playback_process = None

            
        logger.debug("_record_audio_v2")
        while shared_variables.pregong_playing == True :
            time.sleep(0.001)

        #logger.debug(f"pregong playing {shared_variables.pregong_playing}")

        #time.sleep(1)

        try:
            content_length = int(
                self.headers["Content-Length"]
            )  # <--- Gets the size of data
        except Exception as e:
            logger.error(f"Error while getting content: {e}")
            return self._set_response(code=HTTPStatus.CONFLICT.value, message=f"Error while getting content: {e}")

        try:
            message = self.rfile.read(content_length)
            token, audio = extract_token_and_audio(message)
        except Exception as e:
            logger.error(f"Error while reading the content: {e}")
            return self._set_response(code=HTTPStatus.CONFLICT.value, message=f"Error while reading the content: {e}")

        shared_variables.downloading_recorded_message = constants.YES

        try:
            if token == shared_variables.users.secret(shared_variables.active_user):
                
                if shared_variables.modbus_running == True:
                    error_msg = "Modbus running"
                    logger.warning(error_msg)
                    self._set_response(code=HTTPStatus.CONFLICT.value, message=error_msg)
                    return
                
                # Local playback
                try:
                    shared_variables.local_playback_process = Thread(
                        target=shared_variables.write_to_local_playback,
                        args=(audio,),
                        name="AlsaPlayer"
                    )
                    shared_variables.local_playback_process.daemon = True
                    shared_variables.local_playback_process.start()
                except TypeError:
                    logger.error(f"Oops: {sys.exc_info()[0]} occurred")
                except OSError as error:
                    logger.error("While in shared_variables.local_alsa_player.write() the following error occurred:")
                    logger.error(error)
                    logger.error("Check audio device: %s", shared_variables.config["audio_device"]["playback"])

                # sent to groups
                if shared_variables.pmaster.state == 130:
                    data_block_size = 480
                    delay = 0.03
                    logger.info("Sample rate: %s", shared_variables.config["sample_rate"])
                    if shared_variables.config["sample_rate"] == 8000:
                        data_block_size = 912
                        delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"]))
                        # delay = 0.02
                    elif shared_variables.config["sample_rate"] == 24000:
                        data_block_size = 964
                        delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"]))
                        # delay = 0.02
                    else:
                        logger.warning("Sample rate not supported")

                    # First record the message on the disk
                    with open(shared_variables.record_service.current_file, "ab") as message_file:
                        logger.info(f"Writing audio to: {shared_variables.record_service.current_file}")
                        message_file.write(audio)  # PCM 16bit

                    # Notify that the file has been downloaded
                    shared_variables.downloading_recorded_message = constants.NO
                    time.sleep(0.15)  # Give some time to read the flag
                    # Then read from the recorded file
                    with open(shared_variables.record_service.current_file, "rb") as message_file:
                        audio_file = message_file.read()  # PCM 16bit

                        logger.info("Reading from recorded file...")
                        # logger.info(f"Current fast packets value: {shared_variables.first_fast_packets}")
                        retVal = encode_send_audio(audio_file, shared_variables.config["audio_format"],
                                                data_block_size, delay)
                        
                        if retVal == 1:
                            msg = f"Modbus running"
                            logger.error(msg)
                            self._set_response(code=HTTPStatus.CONFLICT.value, message=msg)

                else:
                    logger.warning("Current paging master state: %s", shared_variables.pmaster.state)

                # Kill the local playback process. Important for the case when the user stops the playback earlier
                #if shared_variables.local_playback_process is not None:
                #    shared_variables.kill_alsa_player()
                #    shared_variables.local_playback_process.join(timeout=0.5)
                #    shared_variables.local_playback_process = None

                self._set_response(code=HTTPStatus.NO_CONTENT.value)  # Notify front-end we are done
            else:
                msg = f"Could not authenticate user with ID: {token}"
                logger.error(msg)
                self._set_response(code=HTTPStatus.CONFLICT.value, message=msg)

        except KeyError as e:
            # The request could not be processed because of conflict in the current state of the resource,
            if shared_variables.active_user is None:
                msg = f"Key error. The current user is: {e}"
                logger.error(msg)
            else:
                msg = f"Key error. Message: {e}"
                logger.error(msg)
                traceback.print_exc()
            self._set_response(code=HTTPStatus.CONFLICT.value, message=msg)


    def _get_playfile_state(self):
        """
            While a file is being played, the front end can get the status by sending a /playfile PUT request
        """
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)

        logger.debug(f"pregong_playing {shared_variables.pregong_playing}")

        if shared_variables.modbus_running == True:
            msg = "Modbus running"
            self._set_response(code=HTTPStatus.CONFLICT.value, message=msg)
            return
        

        if not shared_variables.paging_controller._check_user(message["token"]):
            # logger.info(f"User matched: {shared_variables.active_user}")
            error_msg = f"Another user is paging: {shared_variables.active_user}"
            new_user = shared_variables.users.get_user_from_token(message["token"])
            logger.info(f"Received request from {new_user}")
            logger.warning(error_msg)
            front_end_reply = "Another user is paging!"
            self._set_response(code=HTTPStatus.CONFLICT.value, message=front_end_reply)
            return 

        uuid = None

        try:
            uuid = message['uuid']
        except Exception as e:
            logger.error(f"Following occurred: {e}")

        if uuid is None:
            error_msg = "uuid not defined"
            self._set_response(code=HTTPStatus.UNAUTHORIZED.value, message=error_msg)
            return

        if shared_variables.active_user_uuid is not None and uuid != shared_variables.active_user_uuid:
            logger.warning(f"Same user trying to page from a different UUID not allowed!"
                            f"\nRequest -> {message},\nCurrent UUID -> {shared_variables.active_user_uuid}")
            error_msg = "Same user trying to page from a different device not allowed!"
            self._set_response(code=HTTPStatus.CONFLICT.value, message=error_msg)
            return


        #logger.debug("_get_playfile_state"
        #             f"\n\t playfile_state    {shared_variables.playfile_state}"
        #             f"\n\t pregong_playing   {shared_variables.pregong_playing}"
        #             f"\n\t webRequest_playing_file {shared_variables.webRequest_playing_file}"
        #             f"\n\t active_audio_file {shared_variables.audio_files.active_audio_file}"
        #             )



        if shared_variables.playfile_state == constants.BUSY or shared_variables.pregong_playing or shared_variables.webRequest_playing_file:
            msg = "Audio file is currently being played"
            response = shared_variables.audio_files.active_audio_file  # Currently used file, if active
            self._set_response(code=HTTPStatus.PARTIAL_CONTENT.value, message=response)
        elif shared_variables.playfile_state == constants.IDLE:
            msg = "No audio file is currently being played"
            logger.info(msg)
            response = shared_variables.audio_files.active_audio_file  # Currently used file, if active
            self._set_response(code=HTTPStatus.OK.value, message=response)
        else:
            msg = ">>> Unknown playfile state <<<"
            logger.warning(msg)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=msg)

        #if shared_variables.paging_controller._check_user(message["token"]):
        #    if shared_variables.playfile_state == constants.BUSY or shared_variables.pregong_playing:
        #        msg = "Audio file is currently being played"
        #        response = shared_variables.audio_files.active_audio_file  # Currently used file, if active
        #        self._set_response(code=HTTPStatus.PARTIAL_CONTENT.value, message=response)
        #    elif shared_variables.playfile_state == constants.IDLE:
        #        msg = "No audio file is currently being played"
        #        logger.info(msg)
        #        response = shared_variables.audio_files.active_audio_file  # Currently used file, if active
        #        self._set_response(code=HTTPStatus.OK.value, message=response)
        #    else:
        #        msg = ">>> Unknown playfile state <<<"
        #        logger.warning(msg)
        #        self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=msg)
        #else:
        #    logger.warning(f"Internal server error, {message}")
        #    logger.warning("Unable to authenticate user")
        #    # The paging state is not ended correctly
        #    self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message="The paging state is not ended correctly")

#======================================================================================================================
# do_POST
#======================================================================================================================

    def _start_stop_page_message(self):
        if not click_allowed():
            front_end_reply = "Button pressed too fast!"
            logger.warning(front_end_reply)
            self._set_response(code=HTTPStatus.CONFLICT.value, message=front_end_reply)
            return
        set_last_click_time(_stamp())

        shared_variables.downloading_recorded_message = constants.YES
        
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        
        #logger.debug("Message received: %s", message)

        message = json.loads(message)

        if self._check_all_permissions(message) == False:
            return

       
        shared_variables.active_user = shared_variables.users.get_user_from_token(message["token"])
        
        #if message["type"] == 8 and shared_variables.last_message_type == 3:
        if message["type"] == 8:
            logger.debug("STOP paging")
            message["path"] = self.path
            stop_paging_message(message)
            shared_variables.watchDog_live_paging.stop()

            # test - because of pregong
            
            stop_playing_file(message)

            self._set_response(code=HTTPStatus.NO_CONTENT.value)

        elif message["type"] == 3:

            try:

                response_msg = "START paging"
                shared_variables.webRequest_playing_file = True

                logger.debug(response_msg)

                message["path"] = self.path
                
                #start_paging_message(message)
                

                pregong_found, pregong_file = shared_variables.check_pregong()
                if pregong_found:
                    logger.debug("pregong found")
                    
                    shared_variables.pregong_playing = True
                    shared_variables.pregong_message = message.copy()
                    message['file_to_open'] = pregong_file

                    shared_variables.downloading_recorded_message = constants.NO
                    shared_variables.first_fast_packets = 0
                    shared_variables.kill_playfile_thread = constants.NO
                    shared_variables.playfile_state = constants.BUSY
                    start_playing_file(message) # play pregong

                    if self.path == "/recording":
                        #shared_variables.pregong_function = None
                        shared_variables.pregong_function = start_paging_message
                        #start_paging_message(message)

                    elif self.path == "/paging":
                        shared_variables.pregong_function = None
                        
                        #start_paging_message(shared_variables.pregong_message)

                        shared_variables.pregong_function = start_paging_message

                    else:
                        print(f"OTHER {self.path}")
                        logger.error("Requested path unrecognized")

                else:
                    logger.debug("pregong not found")
                    start_paging_message(message)

                self._set_response(code=HTTPStatus.NO_CONTENT.value, message=response_msg)

            except:
                self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message="error ocurred when trying to start paging")

        else:
            logger.error(f"Internal server error:\nMessage -> {message},\ncurrent state ->")
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message="The paging state is not ended correctly")






    # Play a message that has been stored in the memory of the device
    def _start_stop_playing_file(self):
        shared_variables.downloading_recorded_message = constants.NO
        
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)

        #logger.debug(f"Current message from front end: {message}")

        if self._check_all_permissions(message) == False:
            return

        shared_variables.active_user = shared_variables.users.get_user_from_token(message["token"])

        # This is the correct order of the messages when playfile should be STOPPED
        #if message["type"] == 8 and shared_variables.last_message_type == 3:
        if message["type"] == 8:
            logger.debug("STOP playing file")
            stop_playing_file(message)

            self._set_response(code=HTTPStatus.NO_CONTENT.value)

        elif message["type"] == 3:
            logger.debug("START playing file")
            shared_variables.webRequest_playing_file = True

            message['file_to_open'] = self.get_file_to_open(message)
            
            pregong_found, pregong_file = shared_variables.check_pregong()
            if pregong_found:
                logger.debug("pregong found")
                shared_variables.pregong_playing = True
                shared_variables.pregong_function = start_playing_file
                shared_variables.pregong_message = message.copy()
                message['file_to_open'] = pregong_file

                shared_variables.downloading_recorded_message = constants.NO
                shared_variables.first_fast_packets = 0
                shared_variables.kill_playfile_thread = constants.NO
                shared_variables.playfile_state = constants.BUSY
                start_playing_file(message) # play pregong

            else:
                logger.debug("pregong not found")
                start_playing_file(message)

            self._set_response(code=HTTPStatus.NO_CONTENT.value)

        else:
            logger.error(f"Internal server error, {message}")
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message="The playfile state is not ended correctly")
    


    def _create_group(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        # logger.warning("Inside _create_group. Message: %s", message)  # Anton
        try:
            status, reason = shared_variables.groups.add(message)
            if status:
                self.send_response(HTTPStatus.CREATED.value)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            else:
                self._set_response(HTTPStatus.BAD_REQUEST.value)
        except Exception as e:
            logger.warning(f"Create group error")
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.error(message)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=message)

    def _create_audio_file(self):
        # print("Inside _create_audio_file.")  # Anton
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        logger.debug("Message: %s", message)  # Anton
        try:
            status, reason = shared_variables.audio_files.add(message)
            if status:
                self.send_response(HTTPStatus.CREATED.value)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            else:
                self._set_response(HTTPStatus.BAD_REQUEST.value)
        except Exception as e:
            logger.error(f"Create audio file error")
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.error(message)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=message)

    def _create_user(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.users.add(message)
            if status:
                self.send_response(HTTPStatus.CREATED.value)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            else:
                self.send_response(HTTPStatus.BAD_REQUEST.value)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
        except Exception as e:
            logger.error(f"Create user error")
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.error(message)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=message)

#======================================================================================================================
# do_DELETE
#======================================================================================================================

    def _renew_user_token(self):
        m = re.match(r"\/users\/(?P<id>\d+)\/token", self.path)
        user_id = m.groupdict()["id"]
        try:
            status, reason = shared_variables.users.renew_token(user_id)
            if status:
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            elif status == -1:
                self._set_response(404)
            else:
                self._set_response(400)

        except Exception as e:
            msg = f"Renew user token error: {e}"
            logger.error(msg)
            self._set_response(code=500, message=msg)

    def _delete_group(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.groups.remove(message)
            if status:
                self._set_response(HTTPStatus.OK.value)
            else:
                self._set_response(HTTPStatus.NOT_FOUND.value)
        except Exception as e:
            msg = f"Delete group error: {e}"
            logger.error(msg)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=msg)

    def _delete_audio_file(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)

        try:
            status, reason = shared_variables.audio_files.remove(message)
            if status:
                self._set_response(HTTPStatus.OK.value)
            else:
                self._set_response(HTTPStatus.NOT_FOUND.value)
        except Exception as e:
            msg = f"Delete audio file error: {e}"
            logger.error(msg)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=msg)

    def _delete_user(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.users.remove(message)
            if status:
                self._set_response(HTTPStatus.OK.value)
            else:
                self._set_response(HTTPStatus.NOT_FOUND.value)
        except Exception as e:
            msg = f"Delete user error: {e}"
            logger.error(msg)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=msg)


#======================================================================================================================
# do_UPDATE
#======================================================================================================================

    def _update_group(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.groups.update(message)
            if status:
                self.send_response(HTTPStatus.OK.value)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            elif status == -1:
                self._set_response(HTTPStatus.NOT_FOUND.value)
            else:
                self._set_response(HTTPStatus.BAD_REQUEST.value)

        except Exception as e:
            logger.error(f"Create group error")
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.error(message)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=message)

    def _update_audio_file(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.audio_files.update(message)
            if status:
                self.send_response(HTTPStatus.OK.value)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            elif status == -1:
                self._set_response(HTTPStatus.NOT_FOUND.value)
            else:
                self._set_response(HTTPStatus.BAD_REQUEST.value)

        except Exception as e:
            msg = f"Update audio file error: {e}"
            logger.error(msg)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=msg)

    def _update_user(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.users.update(message)
            if status:
                self.send_response(HTTPStatus.OK.value)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            elif status == -1:
                self._set_response(HTTPStatus.NOT_FOUND.value)
            else:
                self._set_response(HTTPStatus.BAD_REQUEST.value)

        except Exception as e:
            msg = f"Create group error: {e}"
            logger.error(msg)
            self._set_response(code=HTTPStatus.INTERNAL_SERVER_ERROR.value, message=msg)


#======================================================================================================================
# END HTTP FUNCTIONS
#======================================================================================================================




global last_click_time

httpd = None
last_click_time = -1

def set_last_click_time(new_time):
    logger.debug("Button click")
    global last_click_time
    if new_time != last_click_time:
        last_click_time = new_time

def click_allowed():
    """
        Check if the CLICK_TIMEOUT has expired, button can be clicked again
        return: True if expired or not yet used, False otherwise
    """

    global last_click_time
    CLICK_TIMEOUT = 1
    
    # If default value, the stream is playing normally and no switching has occured
    if last_click_time == -1:
        logger.debug("Default case, switching normally")
        return True
    now = _stamp()
    diff = now - last_click_time
    if diff > CLICK_TIMEOUT:
        logger.debug(
            f"Timeout reached [{CLICK_TIMEOUT}] -> stream switching can continue normally")
        set_last_click_time(-1)
        return True
    logger.debug(f"Wait for timeout - {diff}")
    return False


def _stamp():
    return datetime.datetime.now().timestamp()


# Method for starting and running  the server
def run_server(server_class=HTTPServer, handler_class=PagingHttpRequestHandler, port: int = 8765,):
    global httpd
    server_address = ("0.0.0.0", port)
    httpd = server_class(server_address, handler_class)
    logger.info("Starting httpd...")
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass

    if httpd is not None:
        httpd.server_close()
    logger.info("Stopping httpd...")


def stop_server():
    # Kill the local playback process. Important for the case when the user stops the playback earlier
    if shared_variables.local_playback_process is not None:
        shared_variables.kill_alsa_player()
        shared_variables.local_playback_process.join(timeout=0.5)
        shared_variables.local_playback_process = None

    shared_variables.playfile_state = constants.IDLE
    shared_variables.audio_files.set_active_audio_file(None)

    # Store current alsa settings, so they can be restored on reboot
    store_alsa_settings()

    global httpd
    if httpd is not None:
        httpd.server_close()
        logger.info("Stopping httpd...")
