#!/usr/bin/env python3
"""Pymodbus Server With Callbacks.

This is an example of adding callbacks to a running modbus server
when a value is written to it.
"""
import asyncio
import logging


from utils import get_logger
from time import sleep
import threading


#from examples.server_async import run_async_server, setup_server
from pymodbus.datastore import (
    ModbusSequentialDataBlock,
    ModbusServerContext,
    ModbusSlaveContext,
)


logger = get_logger("ModbusServer")


from pymodbus.server import StartAsyncTcpServer
from pymodbus.transaction import ModbusSocketFramer
from pymodbus.device import ModbusDeviceIdentification


#
async def run_async_server(context, framer="socket", port=502):
    """Run server."""
    address = ("", port)
    
    txt = f"### start ASYNC server, listening on {address}:{port}"
    logger.info(txt)

    identity = ModbusDeviceIdentification(
        info_name={
            "VendorName": "Pymodbus",
            "ProductCode": "PM",
            "VendorUrl": "https://github.com/pymodbus-dev/pymodbus/",
            "ProductName": "Pymodbus Server",
            "ModelName": "Pymodbus Server",
            "MajorMinorRevision": "3.4.1", #pymodbus_version,
        }
    )

    server = await StartAsyncTcpServer(
        context=context,  # Data storage
        identity=identity,  # server identify
        # TBD host=
        # TBD port=
        address=address,  # listen address
        # custom_functions=[],  # allow custom handling
        framer=ModbusSocketFramer,  # The framer strategy to use
        # ignore_missing_slaves=True,  # ignore request to a missing slave
        # broadcast_enable=False,  # treat slave_id 0 as broadcast address,
        # timeout=1,  # waiting time for request to complete
        # TBD strict=True,  # use strict timing, t1.5 for Modbus RTU
    )
    
    return server





class CallbackDataBlock(ModbusSequentialDataBlock):
    """A datablock that stores the new value in memory,

    and passes the operation to a message queue for further processing.
    """

    def __init__(self, queue, addr, values, callback_play_audio=None):
        """Initialize."""
        logger.info("init CallbackDataBlock")
        self.queue = queue
        self.callback_play_audio = callback_play_audio
        super().__init__(addr, values)

    def setValues(self, address, value):

        """Set the requested values of the datastore."""
        super().setValues(address, value)
        txt = f"Callback from setValues with address {address}, value {value}"
        logger.debug(txt)

        if address == 2:
            if(self.callback_play_audio != None):
                self.callback_play_audio()

    def setValues_no_callback(self, address, value):

        """Set the requested values of the datastore."""
        super().setValues(address, value)


    def getValues(self, address, count=1):
        
        """Return the requested values from the datastore."""
        result = super().getValues(address, count=count)
        txt = f"Callback from getValues with address {address}, count {count}, data {result}"
        logger.debug(txt)
        
        return result

    def validate(self, address, count=1):
        """Check to see if the request is in range."""
        result = super().validate(address, count=count)
        txt = f"Callback from validate with address {address}, count {count}, data {result}"
        logger.debug(txt)
        return result


async def run_callback_server(block=None, cmdline=None, port=502):
    """Define datastore callback for server and do setup."""
 
    store = ModbusSlaveContext(di=None, co=None, hr=block, ir=None)
    context = ModbusServerContext(slaves=store, single=True)

    logger.debug("run ModbusServer")
    await run_async_server(context, port)


if __name__ == "__main__":
    asyncio.run(run_callback_server(), debug=True)  # pragma: no cover


class Modbus:
    def __init__(self, callback_play_audio, port=502):
        logger.info("Modbus initialize...")
        self.port = port
        queue = asyncio.Queue()
        #block = CallbackDataBlock(queue, 0x00, [17] * 100)
        self.block = CallbackDataBlock(queue, 0x00, [0] * 5, callback_play_audio)

        #self.server = ModbusServer("127.0.0.1", 12345, no_block=True, data_bank=icpgw_DataBank(callback_play_audio=callback_play_audio))
        #self.server = ModbusServer("192.168.20.35", 12345, no_block=True, data_bank=icpgw_DataBank(callback_play_audio=callback_play_audio))

        self.th = threading.Thread(target=self.run, args=(), daemon=True)
        self.th.setName("ModbusServer")
        self.th.start()

    def run(self):        
        #self.server.start()
        asyncio.run(run_callback_server(self.block, port=self.port), debug=True)  # pragma: no cover
        #while True:
        #    sleep(1)
        #server.stop()
        logger.debug("@@@@@@@@@@@ exit modbus @@@@@@@@@@@")

