import logging
import logging.config
from logging.handlers import RotatingFileHandler
from datetime import datetime
import pathlib
import json
import socket
from os import system
from subprocess import Popen, PIPE, TimeoutExpired
from inspect import currentframe, getframeinfo

import constants


def get_logger(name="", log_file="/var/log/icpgw", max_file_size=2*1024*1024, config=""):
    if config == "":
        print(f"Creating default logger for {name}")
        new_logger = logging.getLogger(name)
        #new_logger.setLevel(logging.INFO)
        new_logger.setLevel(logging.DEBUG)

        # Initialize directory structure
        log_file = pathlib.Path(log_file)
        log_file.touch(exist_ok=True)
        # ch = logging.FileHandler(log_file)
        ch = RotatingFileHandler(log_file, maxBytes=max_file_size, backupCount=5)
        #ch.setLevel(logging.INFO)
        ch.setLevel(logging.DEBUG)

        # create formatter, to get path: - [%(pathname)s]
        formatter = logging.Formatter(
            "[%(asctime)s " + f"{name: <22}" + "%(lineno)4d %(threadName)16s %(levelname)8s] %(message)s",
            datefmt="%H:%M:%S",
        )

        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger
        new_logger.addHandler(ch)
    else:
        print("About to load .conf file: %s", config)
        logging.config.fileConfig(config)
        new_logger = logging.getLogger(name)
        print("Loaded the .conf file")
    return new_logger


logger = get_logger("UtilsDebugLog")


def get_line_number():
    cf = currentframe()
    return cf.f_back.f_lineno


def get_active_devices(barp, listen_status_socket, pmaster, active_devices):
    while not pmaster.exit_flag:
        try:
            data, address = listen_status_socket.recvfrom(4096)
        except socket.timeout:
            continue
        msg = barp.decode(data)
        if msg[0]["message_type"] == 128:
            active_devices[msg[0]["sender_id"]] = {
                "state": msg[1]["state"],
                "time_stamp": datetime.now(),
            }

        # Removal of devices with stale status.
        current_time = datetime.now()
        for dev_id in list(active_devices.keys()):
            if (current_time - active_devices[dev_id]["time_stamp"]).seconds > 5:
                del active_devices[dev_id]
    print("Status listening stopped!")


def read_json(file_name):
    with open(file_name, "r") as f:
        return json.load(f)


def is_multicast(ip):
    """
    Performs check on ipv4 ip address to determine if it is multicast or not.
    Note: it is assumed that the address is valid.
    Args:
        ip (str)
    Returns:
        (bool) : True if address is multicast and false if it isn't.
    """
    ip_address = [int(i) for i in ip.split(".")]
    if 224 <= ip_address[0] <= 239:
        return True
    else:
        return False


def is_broadcast(multicast_ip, broadcast_ip):
    """
    Performs check on ipv4 ip address to determine if it is broadcast or not.
    Note: it is assumed that the address is valid.
    Args:
        multicast_ip (str)
        broadcast_ip (str)
    Returns:
        (bool) : True if address is broadcast and false if it isn't.
    """
    if multicast_ip == "0.0.0.0":
        return True
    elif multicast_ip == "255.255.255.255":
        return True
    elif multicast_ip == broadcast_ip:
        return True
    else:
        return False


def load_ip_config():
    """
        Creates a file inside the RAM of the device, containing its IP settings.
    Returns:

    """
    # Only used in this function, so no need to have it available for the whole file
    import re

    file_with_path = constants.IF_CONFIG_PATH + constants.IF_CONFIG_FILE
    ip_settings = dict()
    # declaring the regex pattern for IP addresses
    pattern = re.compile('''((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)''')

    # First create the ifconfig file in the IF_CONFIG_PATH
    settings_file = open(file_with_path, "w")
    command = ["ifconfig", "eth0"]
    process = Popen(command, stdout=settings_file, stderr=PIPE, )
    try:
        outs, errs = process.communicate(timeout=5)

        settings_file.close()  # Close the fd after the data has been added to it

        # logger.info(f"Sent command: {command}")
        # logger.info(f"Result from the command: {outs}")

        # Second open the file and extract the ip, netmask and broadcast
        with open(file_with_path, 'r') as f:
            file = f.readlines()
            valid = []
            # Find lines with IPs
            for line in file:
                line = line.rstrip()
                result = pattern.search(line)

                # valid IP addresses
                if result:
                    valid.append(line)
            valid = valid[0].strip().replace("  ", " ").split(" ")  # Clean up the data from unnecessary spaces

            # Store the values to key-value pairs
            for index in range(len(valid)):
                if index % 2 == 0:
                    ip_settings[valid[index]] = valid[index+1]

            # logger.info(f"Valid IP addresses: {valid}")

    except TimeoutExpired:
        process.kill()
        outs, errs = process.communicate()
        print(f"Timed out when loading IP config. Errs :{errs}")

    # logger.info(f"Returning following settings: {ip_settings}")
    return ip_settings


def get_uci(package, section, option):
    command = ["uci", "get", f"{package}.{section}.{option}"]
    output, error = Popen(command, stdout=PIPE, stderr=PIPE).communicate()
    if len(error) == 0:
        return output.decode().strip("\n")
    else:
        raise f"Unable to get {package}.{section}.{option} uci setting : {error}"


def uci_to_config_dict():
    config_dict = {}
    package = "icpgw"

    config_dict["ports"] = {
        "audio": int(get_uci(package, "barp", "audio_port")),
        "control": int(get_uci(package, "barp", "control_port")),
        "status": int(get_uci(package, "barp", "status_port")),
    }
    config_dict["sample_rate"] = int(get_uci(package, "barp", "sample_rate"))
    config_dict["multicast_group_address"] = get_uci(package, "barp", "audio_multicast_group_address")
    config_dict["audio_device"] = {
        "playback": get_uci(package, "server", "playback_device"),
        "record": "plug:analog_out",
    }
    config_dict["encoding_type"] = 0
    config_dict["paging_master_id"] = int(get_uci(package, "barp", "master_id"))
    config_dict["barp_ip"] = get_uci(package, "barp", "ip")
    config_dict["barp_header_kwargs"] = {
        "flags": int(get_uci(package, "barp", "flags")),
        "sender_id": int(get_uci(package, "barp", "master_id")),
        "receiver_id": 0,
        "sequence_number": int(get_uci(package, "barp", "master_id")),
    }
    config_dict["audio_format"] = get_uci(package, "barp", "format")
    config_dict["browser_timeout"] = int(get_uci(package, "server", "timeout"))
    config_dict["paging_priority"] = get_uci(package, "barp", "paging_priority")
    # Returns this kind of dict: {'inet': '192.168.10.65', 'netmask': '255.255.255.0', 'broadcast': '192.168.10.255'}
    config_dict["device_ip_settings"] = load_ip_config()

    config_dict["rtp_port"] = int(get_uci(package, "rtp", "port"))
    config_dict["rtp_ip"] = get_uci(package, "rtp", "ip")
    config_dict["rtp_format"] = get_uci(package, "rtp", "format")
    config_dict["rtp_enabled"] = True if get_uci(package, "rtp", "enabled") == "True" else False

    config_dict["modbus_enabled"] = True if get_uci(package, "modbus", "enabled") == "True" else False
    config_dict["modbus_port"] = int(get_uci(package, "modbus", "port"))

    #config_dict["pregong_enabled"] = True if get_uci(package, "pregong", "enabled") == "True" else False

    return config_dict

def check_uci_pregong():
    package = "icpgw"
    return True if get_uci(package, "pregong", "enabled") == "True" else False

def check_wav_file_format(file_name):
    tmp_file = file_name.split('.')

    if (len(tmp_file) == 2 and tmp_file[1] != "wav") or len(tmp_file) == 1:
        return tmp_file[0] + ".wav"
    return file_name


def get_ip():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        ip = s.getsockname()[0]
        s.close()
        return ip
    except Exception as e:
        print("Couldn't get the ip")
        print(e)
        return "Not found"


def load_alsa_settings(system_func: bool = True):
    logger.info("Loading alsa settings")
    command = [
        'alsactl',
        'restore',
    ]
    if system_func:
        system(" ".join(command))
        return
    return execute_command(command)


def store_alsa_settings(system_func: bool = True):
    logger.info("Storing alsa settings")
    command = [
        'alsactl',
        'store',
    ]
    if system_func:
        system(" ".join(command))
        return
    return execute_command(command)


def execute_command(command: list):
    """
        Open a process to execute the given command
    @param command: List of strings
    @return: True if successful, False otherwise
    """

    process = Popen(
        command,
        stdout=PIPE,
        stderr=PIPE,
    )
    try:
        # logger.info("Executing command: %s", " ".join(command))
        outs, errs = process.communicate(timeout=5)

        if process.returncode != 0:
            logger.error(f"Error occurred: {errs}")
            return False
        # logger.info("Success")
        return True

    except Exception as e:
        logger.warning(f"While trying to execute command: [{command}], the following exception occurred: {e}")
        return False
