import {Component, OnInit, Output, EventEmitter} from '@angular/core';
import {MatDialog, MatDialogConfig} from '@angular/material/dialog';
import {UploadFilesModalComponent} from '../upload-files-modal/upload-files-modal.component';
import {AdditionalCertificatesService} from '../services/additional.certificates.service';
import {HttpErrorResponse} from '@angular/common/http';
import {HTTPResponseHandlerService} from '../services/httpresponse-handler.service';
import {ToastMessageService} from "../services/toast-message.service";
import {SystemInfoService} from "../services/system-info.service";

//import {ToastrService} from 'ngx-toastr';


@Component({
    selector: 'app-additional-certificates',
    templateUrl: './additional.certificates.component.html',
    styleUrls: ['./additional.certificates.component.css', '../css/general.css']
})
export class AdditionalCertificatesComponent implements OnInit {

    constructor(public matDialog: MatDialog,
                private certificatesService: AdditionalCertificatesService,
                private httpResponseHandler: HTTPResponseHandlerService,
                //private toastr: ToastrService,
                private toast: ToastMessageService,
                private systemInfoService: SystemInfoService) {
    }

    displayedColumns = ['select', 'filename', 'status'];
    certificatesList = [];

    availableStorage;

    @Output() loadingEvent = new EventEmitter<boolean>();
    @Output() eventFinished = new EventEmitter<any>();

    ngOnInit(): void {
        //get certificates installed from backend
        this.certificatesService.get().subscribe(data => {
            //console.log("Status:", data.status);
            //console.log("Received:", data.body);
            switch (data.status) {
                case 200:
                    //store for backup
                    //console.log(data.body);
                    this.certificatesList = data.body["custom"];
                    this.systemInfoService.getSystemInfo(['storage_size']).subscribe((data) => {
                        switch (data.status) {
                            case 200:
                                this.availableStorage = data.body["storage_size"]["available"];
                                //console.log(this.storage);
                                break;
                            case 401:
                                this.httpResponseHandler.handle401HttpErrorResponse((data as HttpErrorResponse).error);
                                break;
                            case 500:
                                this.httpResponseHandler.handle500HttpErrorResponse();
                                break;
                            default:
                                console.error("Unexpected return code: ", data.status);
                        }
                    });

                    break;
                case 401:
                    this.httpResponseHandler.handle401HttpErrorResponse((data as HttpErrorResponse).error);
                    break;
                case 500:
                    this.httpResponseHandler.handle500HttpErrorResponse();
                    break;
                default:
                    console.error("Unexpected return code: ", data.status);
            }
        });
    }

    selectOrDeselectCertificate(state: boolean, idx) {
        //console.log(state);
        if (state) {
            this.certificatesList[idx]["selected"] = true;
        } else {
            this.certificatesList[idx]["selected"] = false;
        }
    }

    isAnyCertificateSelected() {
        for (let i = 0; i < this.certificatesList.length; i++) {
            if (this.certificatesList[i]["selected"]) {
                return true;
            }
        }
        return false;
    }

    selectAll() {
        for (let i = 0; i < this.certificatesList.length; i++) {
            this.certificatesList[i]["selected"] = true;
        }
    }

    selectNone() {
        for (let i = 0; i < this.certificatesList.length; i++) {
            this.certificatesList[i]["selected"] = false;
        }
    }

    deleteSelectedCerts() {
        let listOfFilesToBeDeleted = []
        let stringListOfFilesToBeDeleted = "\n• ";

        for (let i = 0; i < this.certificatesList.length; i++) {
            if (this.certificatesList[i]["selected"]) {
                stringListOfFilesToBeDeleted += this.certificatesList[i]["name"] + "\n• ";
                listOfFilesToBeDeleted.push(this.certificatesList[i]["name"]);
            }
        }

        if (listOfFilesToBeDeleted.length === 0) {
            this.toast.showWarningMessage("Please select the certificates to be removed first");
        } else {
            stringListOfFilesToBeDeleted = stringListOfFilesToBeDeleted.substring(0, stringListOfFilesToBeDeleted.lastIndexOf("\n• "));

            if (confirm("Are you sure you want to delete the custom certificates?\n" + stringListOfFilesToBeDeleted)) {
                //delete request to backend
                this.loadingEvent.emit(true);
                this.certificatesService.delete(listOfFilesToBeDeleted).subscribe(res => {
                    switch (res.status) {
                        case 200:
                            if (res.body === "0") {
                                this.eventFinished.emit({
                                    'toast-msg': "No certificates have been removed",
                                    'toast-class': 'my-toast-error'
                                });
                            } else if (res.body === "1") {
                                this.eventFinished.emit({
                                    'toast-msg': "1 certificate was removed",
                                    'toast-class': 'my-toast-success'
                                });
                            } else {
                                this.eventFinished.emit({
                                    'toast-msg': res.body + " certificates were removed",
                                    'toast-class': 'my-toast-success'
                                });
                            }
                            break;
                        case 401:
                            this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                            break;
                        case 500:
                            this.httpResponseHandler.handle500HttpErrorResponse();
                            break;
                        default:
                            console.error("Unexpected return code: ", res.status);
                    }
                });
            }
        }
    }

    openUploadCertificateModal() {
        const dialogConfig = new MatDialogConfig();
        // The user can't close the dialog by clicking outside its body
        dialogConfig.disableClose = false;
        dialogConfig.id = "upload-files-modal-component";
        dialogConfig.height = "200px";
        dialogConfig.width = "900px";
        dialogConfig.position = {left: '6%', top: '10%'}
        dialogConfig.panelClass = "no-padding-modal";
        dialogConfig.data = {
            "component-title": "Upload certificate",
            "button-text": "Upload and install certificate",
            "button-msg": "Upload certificate to install it in the device",
            "toastMsgPositionClass": "",
            "MAX_FILE_SIZE": 16384000,
            "FILE_EXTENSIONS": ".crt",
            "allowMultiple": false,
            "AVAILABLE_STORAGE_SPACE": this.availableStorage,
        }
        //Upload and install certificate
        this.matDialog.open(UploadFilesModalComponent, dialogConfig).afterClosed().subscribe((files) => {
            if (files) {
                //console.log(file);
                this.loadingEvent.emit(true);
                this.certificatesService.upload(files[0]).subscribe(res => {
                    //console.log(res.status);
                    //console.log(res.body);
                    switch (res.status) {
                        case 200:
                            //this.dialogRef.close({"toast-msg" : res.body, "toast-class" : 'my-toast-success'});
                            //console.log(res);
                            this.eventFinished.emit({"toast-msg": res.body, "toast-class": 'my-toast-success'});
                            //console.log("eventFinished emitted");
                            break;
                        case 400:
                            //this.dialogRef.close({"toast-msg" : res.error, "toast-class" : 'my-toast-error'});
                            this.eventFinished.emit({"toast-msg": res.error, "toast-class": 'my-toast-error'});
                            //console.log("eventFinished emitted");
                            break;
                        case 401:
                            this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                            break;
                        case 413:
                            this.eventFinished.emit({
                                "toast-msg": "Error! Maximum upload size exceeded!",
                                "toast-class": 'my-toast-error'
                            });
                            break;
                        case 500:
                            this.httpResponseHandler.handle500HttpErrorResponse();
                            break;
                        case 507:
                            this.eventFinished.emit({
                                "toast-msg": "Device storage is full! Please remove some files to free up some space.",
                                "toast-class": 'my-toast-error'
                            });
                            break;
                        default:
                            console.error("Unexpected return code: ", res.status);
                    }
                });
            }
        });
    }
}
