import {AfterViewInit, Component, OnInit, ViewChild} from '@angular/core';
import {MatTableDataSource} from "@angular/material/table";
import {MatPaginator} from "@angular/material/paginator";
import {HTTPResponseHandlerService} from "../services/httpresponse-handler.service";
import {MatDialog} from "@angular/material/dialog";
import {ToastMessageService} from "../services/toast-message.service";
import {HttpErrorResponse, HttpResponse} from "@angular/common/http";
import {lastValueFrom} from "rxjs";
import {NewAudioFileModalComponent} from "../new-audio-file-modal/new-audio-file-modal.component";
import {AudioFilesService} from "../services/audio-files.service";
import {EditAudioFileModalComponent} from "../edit-audio-file-modal/edit-audio-file-modal.component";
import {Sort} from "@angular/material/sort";
import {compareStrings} from "../utils/comparisons";
import {User} from "../users-management-tab/users-management-tab.component";
import {UsersManagementService} from "../services/users-management.service";

@Component({
  selector: 'app-audio-files-tab',
  templateUrl: './audio-files-tab.component.html',
  styleUrls: ['../css/general.css', './audio-files-tab.component.css']
})
export class AudioFilesTabComponent implements OnInit {

    MAX_N_AUDIO_FILES = 100;

    audioFiles : [];

    all_checked : boolean = false;
    audioFilesSelected = []; // list of ids

    dataIsAvailable: boolean = false;

    /*** Table variables ***/
    displayedColumns = ["selection", "name", "modbus", "timestamp", "action_buttons"];
    dataSource : MatTableDataSource<AudioFile>;

    /*** Paginator variables ***/
    @ViewChild('paginator') paginator: MatPaginator;
    pageIndex : number = 0;
    length : number;
    pageSize : number = 10;
    pageSizeOptions = [5, 10, 25];
    showFirstLastButtons : boolean = true;

    /*** Assign audio files to users ***/
    all_users = [];

    constructor(private usersService: UsersManagementService,
                private audioFilesService: AudioFilesService,
                private httpResponseHandler: HTTPResponseHandlerService,
                public dialog: MatDialog,
                private toast: ToastMessageService) { }

    ngOnInit() {
        this.loadData();
    }

    async loadData() {
        await this.getAllUsers();
        await this.getAudioFiles();
        this.length = this.dataSource.data.length;
        this.audioFilesSelected = [];
        this.all_checked = false;
        this.dataIsAvailable = true;
        setTimeout(() => { //small trick to have the paginator HTML element rendered before assigning it to the table
            this.dataSource.paginator = this.paginator;
        }, 100);
    }

    async getAudioFiles() {
        let audioFiles = [];
        const groupsReq = await this.audioFilesService.getAudioFiles();
        const groupsRes = await lastValueFrom(groupsReq);
        switch (groupsRes.status) {
            case 200:
                this.audioFiles = JSON.parse(groupsRes.body)["audio-files"];
                this.dataSource = new MatTableDataSource(this.audioFiles);
                break;

            case 401:
                this.httpResponseHandler.handle401HttpErrorResponse((groupsRes as HttpErrorResponse).error);
                break;

            case 500:
                this.httpResponseHandler.handle500HttpErrorResponse();
                break;

            default:
                console.error("Unexpected return code: ", groupsRes.status);
        }
        return audioFiles;
    }

    async getAllUsers() {
        this.all_users = [];
        const usersReq = this.usersService.getUsers();
        const usersRes = await lastValueFrom(usersReq);
        switch (usersRes.status) {
            case 200:
                const users = JSON.parse(usersRes.body)["users"];
                users.forEach(user => {
                    if (user["id"] !== "0"){
                        this.all_users.push(user);
                    }
                });
                break;

            case 401:
                this.httpResponseHandler.handle401HttpErrorResponse((usersRes as HttpErrorResponse).error);
                break;

            case 500:
                this.httpResponseHandler.handle500HttpErrorResponse();
                break;

            default:
                console.error("Unexpected return code: ", usersRes.status);
        }
    }

    getAssignedUsers(audio_file_id: string) {
        const assigned_users = [];
        this.all_users.forEach( (user: User) => {
            if (user.id !== "0" && user["audio_files"] && user["audio_files"].includes(audio_file_id)){
                assigned_users.push(user);
            }
        });
        return assigned_users;
    }

    /*** Selection ***/

    audioFileIsSelected(audioFileId){
        return this.audioFilesSelected.indexOf(audioFileId) > -1;
    }

    selectOrDeselectAudioFile(audioFileId){
        const idx = this.audioFilesSelected.indexOf(audioFileId);
        if(idx > -1){
            this.audioFilesSelected.splice(idx, 1);
        } else {
            if (audioFileId !== '0')
                this.audioFilesSelected.push(audioFileId);
        }
        this.all_checked = this.audioFilesSelected.length === this.dataSource.data.length;
    }

    selectOrDeselectAll(){
        if (this.all_checked){
            this.audioFilesSelected = [];
            this.all_checked = false;
        } else {
            // select all
            this.dataSource.data.forEach(groupElem => {
                if (!this.audioFileIsSelected(groupElem.id) && groupElem.id !== '0') this.audioFilesSelected.push(groupElem.id);
            });
            this.all_checked = true;
        }
    }

    /*** End selection ***/
    createNewAudioFile() {
        const dialogRef = this.dialog.open(NewAudioFileModalComponent, {
            width: "45vw",
            height: "60vh",
            disableClose: true,
            data: {
                all_users: this.all_users,
                modbus_idx_list: this.getModbusIdxInUse(),
                btn_names_list: this.getButtonNamesInUse()
            }
        });

        dialogRef.afterClosed().subscribe(result => {
            if (result){
                this.toast.showSuccessMessage("New device message created successfully!");
                this.dataIsAvailable = false;
                this.loadData();
            } else {
                console.log("canceled");
            }
        });
    }

    async deleteSelectedAudioFiles() {
        if (this.audioFilesSelected.length === 0){
            alert("Please select one or more device messages to delete them.");
        } else {
            let confirm_msg = "Are you sure you want to delete the selected device messages?";
            if (this.anySelectedAudioFileInUse()){
                confirm_msg = "Are you sure you want to delete the selected device messages? Some of the device messages are in use and will not be deleted.";
            }
            if (confirm(confirm_msg) === true) {
                const req = this.audioFilesService.deleteAudioFiles(this.audioFilesSelected);
                const res = await lastValueFrom(req);
                switch (res.status) {
                    case 200:
                        // reload data
                        this.toast.showSuccessMessage("Audio files deleted successfully!");
                        this.dataIsAvailable = false;
                        this.loadData();
                        break;
                    case 401:
                        this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                        break;
                    case 500:
                        this.httpResponseHandler.handle500HttpErrorResponse();
                        break;
                    default:
                        console.error("Unexpected return code: ", res.status);
                }
            }
        }
    }

    editGroup(audioFile: AudioFile) {
        const dialogRef = this.dialog.open(EditAudioFileModalComponent, {
            width: "45vw",
            height: "60vh",
            disableClose: true,
            data: {
                audio_file: audioFile,
                all_users: this.all_users,
                assigned_users: this.getAssignedUsers(audioFile.id),
                modbus_idx_list: this.getModbusIdxInUse(),
                btn_names_list: this.getButtonNamesInUse()
            }
        });

        dialogRef.afterClosed().subscribe(async result => {
            if (result){
                this.toast.showSuccessMessage("Device message updated successfully!");
                this.dataIsAvailable = false;
                this.loadData();
            } else {
                console.log("canceled");
            }

        });
    }

    /*** Sort ***/
    sortData(sort: Sort) {
        const data = this.audioFiles.slice();
        if (!sort.active || sort.direction === '') {
            this.dataSource.data = data;
            return;
        }

        this.dataSource.data = data.sort((a: AudioFile, b:AudioFile) => {
            const isAsc = sort.direction === 'asc';
            switch (sort.active) {
                case 'name':
                    return compareStrings(a.btn_name, b.btn_name, isAsc);
                default:
                    return 0;
            }
        });
    }

    anySelectedAudioFileInUse() {
        let inUse = false;
        for (const audio_file_id of this.audioFilesSelected) {
            for (const user of this.all_users) {
                if (user["audio_files"] && user["audio_files"].includes(audio_file_id)) {
                    inUse = true;
                    break;
                }
            }
            if (inUse){
                break;
            }
        }
        return inUse;
    }

    getModbusIdxInUse() {
        let modbus_idx_list = [];
        for (const audio_file of this.audioFiles){
            if (audio_file["modbus_idx"]){
                modbus_idx_list.push(audio_file["modbus_idx"]);
            }
        }
        return modbus_idx_list;
    }

    getButtonNamesInUse() {
        let btn_names_list = [];
        for (const audio_file of this.audioFiles){
            btn_names_list.push(audio_file["btn_name"]);
        }
        return btn_names_list;
    }

}

export interface AudioFile {
    id: string;
    btn_name: string;
    file_name: string;
    timestamp: Date;
    sample_rate: number;
    modbus_idx: number;
}
