import {Component, Inject, OnInit} from '@angular/core';
import {FormControl, FormGroup, Validators} from "@angular/forms";
import {MAT_DIALOG_DATA, MatDialogRef} from "@angular/material/dialog";
import {ToastMessageService} from "../services/toast-message.service";
import {AudioFilesService} from "../services/audio-files.service";
import {HTTPResponseHandlerService} from "../services/httpresponse-handler.service";
import {HttpErrorResponse, HttpEventType, HttpResponse} from "@angular/common/http";
import {User} from "../users-management-tab/users-management-tab.component";
import {
    ButtonNameCustomValidator,
    ModbusIndexCustomValidator,
} from "../audio-files-tab/device-message-custom-validator.directive";

@Component({
  selector: 'app-edit-audio-file-modal',
  templateUrl: './edit-audio-file-modal.component.html',
  styleUrls: ['../css/general.css','../css/new-or-edit-modals.css', './edit-audio-file-modal.component.css']
})
export class EditAudioFileModalComponent implements OnInit {

    title : string;
    audioFileForm: FormGroup;
    audioFile: File;
    audioFileDetailsForm: FormGroup;

    uploading: boolean = false;
    audioFileUploadProgress: number;

    /*** Assign to users ***/
    assigned_users = [];

    constructor(public dialogRef: MatDialogRef<EditAudioFileModalComponent>,
                @Inject(MAT_DIALOG_DATA) public data,
                private toast: ToastMessageService,
                private audioFilesService: AudioFilesService,
                private httpResponseHandler: HTTPResponseHandlerService) { }

    ngOnInit(): void {
        console.log("[ngOnInit] modbus_idx_list", this.data["modbus_idx_list"]);
        this.title = "Edit Device Message";
        this.audioFileForm = new FormGroup<any>({
            "file": new FormControl<File>(null)
        });
        this.audioFileDetailsForm = new FormGroup<any>({
            "name": new FormControl<string>(this.data.audio_file.btn_name, [Validators.required, ButtonNameCustomValidator(this.data.audio_file.btn_name, this.data["btn_names_list"])]),
            "modbus_idx": new FormControl<number>(this.data.audio_file.modbus_idx ? this.data.audio_file.modbus_idx : null, ModbusIndexCustomValidator(this.data.audio_file.modbus_idx, this.data["modbus_idx_list"])),
        });
        this.assigned_users = this.data.assigned_users;
    }

    onFileChange(event) {
        if (event.target.files.length > 0) {
            this.audioFile = event.target.files[0];
        }
    }

    submit(){
        if (this.audioFileForm.valid && this.audioFileDetailsForm.valid){
            let formData = new FormData();
            if (this.audioFile){
                this.uploading = true;
                formData.append('audio_file', this.audioFile, this.audioFile.name);
            }
            const data = {
                "name": this.audioFileDetailsForm.value['name'],
                "modbus_idx": this.audioFileDetailsForm.value['modbus_idx'] === "" ? null : this.audioFileDetailsForm.value['modbus_idx'],
                "filename": this.audioFile ? this.audioFile.name : this.data.audio_file.file_name,
                "id": this.data.audio_file.id,
                "users": this.getUserIds()
            }

            formData.append('data', JSON.stringify(data));

            this.audioFilesService.saveAudioFileChanges(formData).subscribe(res => {
                if (res.type === HttpEventType.UploadProgress) {
                    const progress = Math.round(100 * res.loaded / res.total);
                    if (progress < 100){
                        this.audioFileUploadProgress = progress;
                    }
                } else if (res instanceof HttpResponse) {
                    switch (res.status) {
                        case 200:
                            this.audioFileUploadProgress = 100;
                            setTimeout(() => {
                                this.dialogRef.close(true);
                            }, 1000);
                            break;
                        default:
                            console.error("Unexpected return code: ", res.status);
                            this.toast.showErrorMessage(res.body);
                            this.dialogRef.close(false);
                    }
                } else if (res instanceof HttpErrorResponse) {
                    console.error("HTTP Error Response" + res);
                    if (res.status === 401) {
                        this.httpResponseHandler.handle401HttpErrorResponse(res.error);
                    } else {
                        this.toast.showErrorMessage((res as HttpErrorResponse).error);
                    }
                    this.dialogRef.close(false);
                }
            }, err => {
                console.error(err);
                if (err.status === 413 || err.statusText === "REQUEST ENTITY TOO LARGE") {
                    this.toast.showWarningMessage("File is too large!");
                } else if (err.error === "No file") {
                    this.toast.showErrorMessage("No file provided");
                } else if (err.status === 401) {
                    this.httpResponseHandler.handle401HttpErrorResponse(err.error);
                } else {
                    this.toast.showErrorMessage(err.error ? err.error : "Error saving device message changes.");
                    this.dialogRef.close(false);
                }
            });
        } else {
            this.toast.showErrorMessage("Audio file data is invalid!");
        }
    }

    getUserIds() {
        const assigned_user_ids = [];
        this.assigned_users.forEach( (user: User) => {
            assigned_user_ids.push(user.id);
        });
        return assigned_user_ids;
    }

    onUsersSelectedChange(selectedUsers: User[]) {
        this.assigned_users = selectedUsers;
    }

    getUsersSelectedNames() {
        let strNames = "";
        this.assigned_users.forEach( (user: User) => {
            strNames += "" + user.name + ", ";
        });
        return strNames.substring(0, strNames.length-2);
    }

}
