import { Component, EventEmitter, HostListener, Inject, Input, OnDestroy, OnInit, Output } from '@angular/core';
import {HttpErrorResponse, HttpEventType, HttpResponse} from '@angular/common/http';
import {FirmwareUploadService} from '../services/firmware-upload.service';
import {HTTPResponseHandlerService} from '../services/httpresponse-handler.service';
import {ToastrService} from 'ngx-toastr';
import {ProgressBarMode} from '@angular/material/progress-bar';
import {SystemFunctionsService} from "../services/system-functions.service";
import {ToastMessageService} from "../services/toast-message.service";
import {FormCanDeactivate} from "../can-deactivate/form-can-deactivate";
import { MAT_DIALOG_DATA, MatDialog, MatDialogRef } from '@angular/material/dialog';


@Component({
    selector: 'app-fw-updater-status-modal',
    templateUrl: './fw-updater-status-modal.component.html',
    styleUrls: ['./fw-updater-status-modal.component.css', '../css/general.css']
})


export class FwUpdaterStatusModalComponent implements OnInit {

    // @Output() updateStatusEvent = new EventEmitter<string>();

    updateStatus : string;

    uploadSubscription;
    progressMessagesStatus = [];
    successColor = '#333333';
    errorColor = 'red';

    progress: number = 0;
    showFileUploadProgressBar: boolean = false;
    showProgressBar: boolean;
    error: boolean = false;

    allowCancel : boolean = true;


    constructor(private fileUploadService: FirmwareUploadService,
                private httpResponseHandler: HTTPResponseHandlerService,
                private toast: ToastMessageService,
                private systemFuncsService: SystemFunctionsService,
                public dialogRef: MatDialogRef<FwUpdaterStatusModalComponent>,
                @Inject(MAT_DIALOG_DATA) public data: File) {}

    @HostListener('window:beforeunload', ['$event'])
    confirmClosing($event: any) {
        // console.log("updateStatus", this.updateStatus);
        // console.log("allowCancel", this.allowCancel);
        // console.log(this.updateStatus === "uploading" && !this.allowCancel);
        // console.log(this.updateStatus !== "uploading" && this.updateStatus !== "error" && this.updateStatus !== "rebooting" && this.updateStatus !== "finished");
        if ( (this.updateStatus === "uploading" && !this.allowCancel) || ( this.updateStatus !== "uploading" && this.updateStatus !== "error" && this.updateStatus !== "rebooting" && this.updateStatus !== "finished") ) {
            // alert("You can not leave the page until the update is finished.");
            return false;
        } else {
           return true;
        }
    }


    ngOnInit(): void {
        this.progressMessagesStatus.push(['Uploading firmware update file "' + this.data.name + '"...', this.successColor]);
        this.showFileUploadProgressBar = true;
        this.updateStatus = "uploading";
        // this.updateStatusEvent.emit(this.updateStatus);
        this.uploadSubscription = this.fileUploadService.upload(this.data).subscribe(res => {
            if (res.type === HttpEventType.UploadProgress) {
                this.progress = Math.round(100 * res.loaded / res.total);
                if (this.progress >= 90) {
                    this.allowCancel = false;
                }
                if (this.progress === 100) {
                    setTimeout(() => {
                        this.showFileUploadProgressBar = false;
                        if (!this.error) this.showProgressBar = true;
                    }, 1000);
                }
            } else if (res instanceof HttpResponse) {
                switch (res.status) {
                    case 200:
                        this.showProgressBar = false;
                        this.progressMessagesStatus.push(['Firmware update file uploaded with success.', this.successColor]);
                        setTimeout(() => {
                            this.progressMessagesStatus.push(['Installing firmware update...', this.successColor]);
                            this.showProgressBar = true;
                            this.checkUpdateStatus();
                        }, 800);
                        break;
                    default:
                        console.error("Unexpected return code: ", res.status);
                }
            } else if (res instanceof HttpErrorResponse) {
                console.error("HTTP Error Response" + res);
                // this.updateStatus = "error";
                // this.updateStatusEvent.emit(this.updateStatus);
                this.error = true;
                this.progressMessagesStatus.push(['Error uploading file. ', this.errorColor]);
                this.showFileUploadProgressBar = false;
            }
        }, err => {
            console.error(err);
            // this.updateStatus = "error";
            // this.updateStatusEvent.emit(this.updateStatus);
            this.error = true;
            this.progressMessagesStatus.push(['Error uploading file. ', this.errorColor]);
            this.showFileUploadProgressBar = false;
            this.showProgressBar = false;
            if (err.status === 413 || err.statusText === "REQUEST ENTITY TOO LARGE") {
                this.toast.showWarningMessage("File is too large!");
            } else if (err.status === 409 || err.statusText === "CONFLICT") {
                this.toast.showErrorMessage("There is an update already running.");
            } else if (err.error === "No file") {
                this.toast.showErrorMessage("No file provided");
            }
        });
    }


    checkUpdateStatus() {
        this.fileUploadService.getUpdateStatus().subscribe(status => {
            if (status.status === 200) {
                if (status.body === "rebooting") {
                    this.showProgressBar = false;
                    this.progressMessagesStatus.push(['Update complete!', this.successColor]);
                    this.updateStatus = "rebooting";
                    // this.updateStatusEvent.emit(this.updateStatus);
                    setTimeout(() => { //to allow user to see
                        this.progressMessagesStatus.push(['Please wait while the device is rebooting. The page will be automatically reloaded.', this.successColor]);
                        setTimeout(() => this.reloadWebUI(), 10000);
                    }, 1000);
                } else if (status.body === "error") {
                    //console.log("Stop checking for status");
                    this.showProgressBar = false;
                    this.showFileUploadProgressBar = false;
                    this.progressMessagesStatus.push(['An error occurred during the installation.', this.errorColor]);
                    this.updateStatus = "error";
                    // this.updateStatusEvent.emit(this.updateStatus);
                } else if (status.body === "finished") {
                    //missed reboot
                    this.updateStatus = "finished";
                    // this.updateStatusEvent.emit(this.updateStatus);
                    this.reloadWebUI();
                } else { //installing
                    this.updateStatus = status.body;
                    // this.updateStatusEvent.emit(this.updateStatus);
                    setTimeout(() => {
                        this.checkUpdateStatus();
                    }, 3000);
                }
            } else if (status.status === 401) {
                // this.updateStatus = "error";
                // this.updateStatusEvent.emit(this.updateStatus);
                this.httpResponseHandler.handle401HttpErrorResponse(status.error);
            } else {
                setTimeout(() => {
                    this.checkUpdateStatus();
                }, 1000);
            }
        });
    }


    reloadWebUI() {
        this.systemFuncsService.appIsRunning().subscribe(res => {
            if (res.status === 200) {
                window.location.href = window.location.protocol + '//' + window.location.hostname;
            } else {
                setTimeout(() => this.reloadWebUI(), 1000);
            }
        });
    }


    cancelUpload(){
        this.uploadSubscription.unsubscribe();
        this.closeModal();
    }


    closeModal() {
        this.dialogRef.close();
    }

}
