import {Component, OnInit, ViewChild} from '@angular/core';
import {MatTableDataSource} from "@angular/material/table";
import {MatPaginator} from "@angular/material/paginator";
import {HTTPResponseHandlerService} from "../services/httpresponse-handler.service";
import {MatDialog} from "@angular/material/dialog";
import {ToastMessageService} from "../services/toast-message.service";
import {HttpErrorResponse} from "@angular/common/http";
import {lastValueFrom} from "rxjs";
import {NewOrEditGroupModalComponent} from "../new-or-edit-group-modal/new-or-edit-group-modal.component";
import {GroupsManagementService} from "../services/groups-management.service";
import {Sort} from "@angular/material/sort";
import {compareStrings} from "../utils/comparisons";

@Component({
  selector: 'app-groups-tab',
  templateUrl: './groups-tab.component.html',
  styleUrls: ['../css/general.css', './groups-tab.component.css']
})
export class GroupsTabComponent implements OnInit {

    MAX_N_GROUPS = 128;

    groupList : [];

    all_checked : boolean = false;
    groupsSelected = []; // list of ids

    dataIsAvailable: boolean = false;

    /*** Table variables ***/
    displayedColumns = ["selection", "name", "id", "action_buttons"];
    dataSource: MatTableDataSource<Group>;

    /*** Paginator variables ***/
    @ViewChild('paginator') paginator: MatPaginator;
    pageIndex : number = 0;
    length : number;
    pageSize : number = 10;
    pageSizeOptions = [5, 10, 25];
    showFirstLastButtons : boolean = true;

  constructor(private groupsService: GroupsManagementService,
              private httpResponseHandler: HTTPResponseHandlerService,
              public dialog: MatDialog,
              private toast: ToastMessageService) { }

    ngOnInit() {
        this.loadData();
    }

    async loadData() {
        await this.getGroups();
        this.length = this.dataSource.data.length;
        this.groupsSelected = [];
        this.all_checked = false;
        this.dataIsAvailable = true;
        setTimeout(() => { //small trick to have the paginator HTML element rendered before assigning it to the table
            this.dataSource.paginator = this.paginator;
        }, 100);
    }

    async getGroups() {
        const groupsReq = this.groupsService.getGroups();
        const groupsRes = await lastValueFrom(groupsReq);
        switch (groupsRes.status) {
            case 200:
                // this.dataSource = new MatTableDataSource(JSON.parse(groupsRes.body)["groups"]);
                this.groupList = JSON.parse(groupsRes.body)["groups"];
                this.dataSource = new MatTableDataSource<Group>(this.groupList);
                break;

            case 401:
                this.httpResponseHandler.handle401HttpErrorResponse((groupsRes as HttpErrorResponse).error);
                break;

            case 500:
                this.httpResponseHandler.handle500HttpErrorResponse();
                break;

            default:
                console.error("Unexpected return code: ", groupsRes.status);
        }
    }

    /*** Selection ***/

    groupIsSelected(userId){
        return this.groupsSelected.indexOf(userId) > -1;
    }

    selectOrDeselectGroup(groupId){
        const idx = this.groupsSelected.indexOf(groupId);
        if(idx > -1){
            this.groupsSelected.splice(idx, 1);
        } else {
            if (groupId !== '0')
                this.groupsSelected.push(groupId);
        }
        this.all_checked = this.groupsSelected.length === this.dataSource.data.length;
    }

    selectOrDeselectAll(){
        if (this.all_checked){
            this.groupsSelected = [];
            this.all_checked = false;
        } else {
            // select all
            this.dataSource.data.forEach(groupElem => {
                if (!this.groupIsSelected(groupElem.id) && groupElem.id !== '0') this.groupsSelected.push(groupElem.id);
            });
            this.all_checked = true;
        }
    }

    /*** End selection ***/

    openNewOrEditGroupDialog(data){
        return this.dialog.open(NewOrEditGroupModalComponent, {
            width: "35vw",
            height: "72vh",
            data: data
        });
    }

    createNewGroup() {
        const data = {
            "mode": "new"
        }
        const dialogRef = this.openNewOrEditGroupDialog(data);

        dialogRef.afterClosed().subscribe(async result => {
            if (result){
                // send request to save new user
                const req = this.groupsService.saveNewGroup(result);
                const res = await lastValueFrom(req);
                switch (res.status) {
                    case 201:
                        // reload data
                        this.dataIsAvailable = false;
                        this.loadData();
                        break;
                    case 400:
                        this.toast.showErrorMessage((res as HttpErrorResponse).error);
                        break;
                    case 401:
                        this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                        break;
                    case 500:
                        this.httpResponseHandler.handle500HttpErrorResponse();
                        break;
                    default:
                        console.error("Unexpected return code: ", res.status);
                }
            } else {
                console.log("canceled");
            }
        });
    }

    async deleteSelectedGroups() {
        if (this.groupsSelected.length === 0){
            alert("Please select one or more users to delete them.");
        } else {
            if (confirm("Are you sure you want to delete the selected users?") === true) {
                const req = this.groupsService.deleteGroups(this.groupsSelected);
                const res = await lastValueFrom(req);
                switch (res.status) {
                    case 200:
                        // reload data
                        this.dataIsAvailable = false;
                        this.loadData();
                        break;
                    case 401:
                        this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                        break;
                    case 500:
                        this.httpResponseHandler.handle500HttpErrorResponse();
                        break;
                    default:
                        console.error("Unexpected return code: ", res.status);
                }
            }
        }
    }

    editGroup(group) {
        const data = {
            "mode": "edit",
            "name": group.name,
            "clients": JSON.parse(JSON.stringify(group.clients))
        }
        const dialogRef = this.openNewOrEditGroupDialog(data);

        dialogRef.afterClosed().subscribe(async result => {
            if (result){
                result["id"] = group.id;
                // send request to save new user
                const req = this.groupsService.saveGroupChanges(result);
                const res = await lastValueFrom(req);
                switch (res.status) {
                    case 200:
                        // reload data
                        this.dataIsAvailable = false;
                        this.loadData();
                        break;
                    case 400:
                        this.toast.showErrorMessage((res as HttpErrorResponse).error);
                        break;
                    case 401:
                        this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                        break;
                    case 500:
                        this.httpResponseHandler.handle500HttpErrorResponse();
                        break;
                    default:
                        console.error("Unexpected return code: ", res.status);
                }
            } else {
                console.log("canceled");
            }

        });
    }

    /*** Sort ***/
    sortData(sort: Sort) {
        const data = this.groupList.slice();
        if (!sort.active || sort.direction === '') {
            this.dataSource.data = data;
            return;
        }

        this.dataSource.data = data.sort((a: Group, b:Group) => {
            const isAsc = sort.direction === 'asc';
            switch (sort.active) {
                case 'name':
                    return compareStrings(a.name, b.name, isAsc);
                case 'id':
                    return compareStrings(a.name, b.name, isAsc);
                default:
                    return 0;
            }
        });
    }
}



export interface Group {
    id: string;
    name: string;
    clients: number[];
}
