import {Component, Inject, OnInit} from '@angular/core';
import {FormControl, FormGroup, Validators} from "@angular/forms";
import {MAT_DIALOG_DATA, MatDialogRef} from "@angular/material/dialog";
import {ToastMessageService} from "../services/toast-message.service";
import {AudioFilesService} from "../services/audio-files.service";
import {HTTPResponseHandlerService} from "../services/httpresponse-handler.service";
import {HttpErrorResponse, HttpEventType, HttpResponse} from "@angular/common/http";
import {Subscription} from "rxjs";
import {User} from "../users-management-tab/users-management-tab.component";
import {
    ButtonNameCustomValidator,
    ModbusIndexCustomValidator
} from "../audio-files-tab/device-message-custom-validator.directive";

@Component({
  selector: 'app-new-or-edit-audio-file-modal-component',
  templateUrl: './new-audio-file-modal.component.html',
  styleUrls: ['../css/general.css', '../css/new-or-edit-modals.css', './new-audio-file-modal.component.css']
})
export class NewAudioFileModalComponent implements OnInit {

    title : string;
    audioFileForm: FormGroup;
    audioFile: File;
    audioFileDetailsForm: FormGroup;

    uploading: boolean = false;
    audioFileUploadProgress: number;

    /*** Assign to users ***/
    assigned_users = [];

  constructor(public dialogRef: MatDialogRef<NewAudioFileModalComponent>,
              @Inject(MAT_DIALOG_DATA) public data,
              private toast: ToastMessageService,
              private audioFilesService: AudioFilesService,
              private httpResponseHandler: HTTPResponseHandlerService) { }

  ngOnInit(): void {
      this.title = "Upload new Device Message";
      this.audioFileForm = new FormGroup<any>({
          "file": new FormControl<File>(null, Validators.required)
      });
      this.audioFileDetailsForm = new FormGroup<any>({
          "name": new FormControl<string>('', [Validators.required, ButtonNameCustomValidator(null, this.data["btn_names_list"])]),
          "modbus_idx": new FormControl<number>(null, ModbusIndexCustomValidator(null, this.data["modbus_idx_list"])),
      });
      this.audioFileDetailsForm.markAllAsTouched();
  }

    onFileChange(event) {
        if (event.target.files.length > 0) {
            this.audioFile = event.target.files[0];
            this.audioFileDetailsForm.patchValue({"name": this.audioFile.name});
        }
    }

  submit(){
    if (this.audioFileForm.valid && this.audioFileDetailsForm.valid){
      this.uploading = true;
      let formData = new FormData();
      formData.append('audio_file', this.audioFile, this.audioFile.name);

      const data = this.audioFileDetailsForm.value;
      if (data["modbus_idx"] === "") {
          data["modbus_idx"] = null;
      }
      data["users"] = this.getUserIds();

      formData.append('data', JSON.stringify(data));

          this.audioFilesService.saveNewAudioFile(formData).subscribe(res => {
              if (res.type === HttpEventType.UploadProgress) {
                  const progress = Math.round(100 * res.loaded / res.total);
                  if (progress < 100){
                      this.audioFileUploadProgress = progress;
                  }
              } else if (res instanceof HttpResponse) {
                  switch (res.status) {
                      case 201:
                          this.audioFileUploadProgress = 100;
                          setTimeout(() => {
                              this.dialogRef.close(true);
                          }, 1000);
                          break;
                      default:
                          console.error("Unexpected return code: ", res.status);
                          this.dialogRef.close(false);
                  }
              } else if (res instanceof HttpErrorResponse) {
                  console.error("HTTP Error Response" + res);
                  if (res.status === 401) {
                      this.httpResponseHandler.handle401HttpErrorResponse(res.error);
                  } else {
                      this.toast.showErrorMessage((res as HttpErrorResponse).error);
                  }
                  this.dialogRef.close(false);
              }
          }, err => {
              if (err.status === 413 || err.statusText === "REQUEST ENTITY TOO LARGE") {
                  this.toast.showErrorMessage("File is too large!");
                  this.dialogRef.close(false);
              } else if (err.error === "No file") {
                  this.toast.showErrorMessage("No file provided");
                  this.dialogRef.close(false);
              } else if (err.status === 401) {
                  this.httpResponseHandler.handle401HttpErrorResponse(err.error);
              } else {
                  this.toast.showErrorMessage("Error saving new device message.");
                  this.dialogRef.close(false);
              }
        });
      } else {
          this.toast.showErrorMessage("Device message data is invalid!");
      }
  }

    getUserIds() {
        const assigned_user_ids = [];
        this.assigned_users.forEach( (user: User) => {
            assigned_user_ids.push(user.id);
        });
        return assigned_user_ids;
    }

    onUsersSelectedChange(selectedUsers: User[]) {
        this.assigned_users = selectedUsers;
    }

    getUsersSelectedNames() {
        let strNames = "";
        this.assigned_users.forEach( (user: User) => {
            strNames += "" + user.name + ", ";
        });
        return strNames.substring(0, strNames.length-2);
    }
}
