import {Component, ElementRef, OnInit} from '@angular/core';
import {FormControl, FormGroup, Validators} from "@angular/forms";
import {buildSettingsFormGroup} from "../settings/settings-utils";
import {HttpErrorResponse} from "@angular/common/http";
import {MatSliderChange} from "@angular/material/slider";
import {DeviceConfigService} from "../services/device-config.service";
import {ToastrService} from "ngx-toastr";
import {ToastMessageService} from "../services/toast-message.service";
import {HTTPResponseHandlerService} from "../services/httpresponse-handler.service";
import {SystemInfoService} from "../services/system-info.service";
import {TopBarService} from "../services/top-bar.service";
import {SystemFunctionsService} from "../services/system-functions.service";
import {MatDialog, MatDialogConfig} from "@angular/material/dialog";
import {FormCanDeactivate} from "../can-deactivate/form-can-deactivate";
import {UploadFilesModalComponent} from "../upload-files-modal/upload-files-modal.component";
import {AudioFilesService} from "../services/audio-files.service";

@Component({
  selector: 'app-paging-settings-tab',
  templateUrl: './paging-settings-tab.component.html',
  styleUrls: ['../css/general.css', '../css/mat-form-field.css', './paging-settings-tab.component.css']
})
export class PagingSettingsTabComponent extends FormCanDeactivate implements OnInit {

    /* Forms */
    settingsForm = new FormGroup<any>({});
    /* Backup */ //used to check what changed exactly, to decide which message should show while submitting
    backupSettingsDict = {};

    ucisList = {
        'barp_settings': 'icpgw.barp',
        // 'rtp_settings': 'icpgw.rtp',
        'icpgw_server_timeout':'icpgw.server.timeout',
        'modbus_settings': 'icpgw.modbus',
        'pregong_enabled': 'icpgw.pregong.enabled'
    };

    dataIsAvailable: boolean = false;
    show_loading_component: boolean = false;
    show_settings_component: boolean = false;

    /* State Variables */

    // submitting status
    rebooting: boolean = false;
    rebooting_staticIPAddress: boolean = false;
    web_proto_changed: boolean = false;

    PAGING_SAMPLE_RATES = [
        {value: 8000, viewValue: '8KHz'},
        {value: 24000, viewValue: '24KHz'}
    ];
    PAGING_AUDIO_FORMATS = [
        {value: 'ulaw', viewValue: '\u03BC-law'},
        {value: 'alaw', viewValue: 'a-law'}
    ];
    RTP_AUDIO_FORMATS = [
        {value: 'PCM16BE', viewValue: 'S16BE'},
        {value: 'G711_ULAW', viewValue: '\u03BC-law'},
        {value: 'G711_ALAW', viewValue: 'a-law'}
    ];

    pregong_filename : string = "N/A";

  constructor(private configsService: DeviceConfigService,
              private toastr: ToastrService,
              private toast: ToastMessageService,
              private httpResponseHandler: HTTPResponseHandlerService,
              private systemInfoService: SystemInfoService,
              private topBarService: TopBarService,
              private systemFuncsService: SystemFunctionsService,
              public matDialog: MatDialog,
              private elem: ElementRef,
              private audioFilesService: AudioFilesService) {
      super();
  }

    canDeactivateComponent() {
        return !this.settingsForm.dirty;
    };


    ngOnInit() {
        this.toast.clearToastMessages();
        this.topBarService.updateDeviceLocalTime();
        this.loadDataFromDevice();
    }


    async loadDataFromDevice() {
        this.dataIsAvailable = false;
        this.settingsForm = new FormGroup({});
        this.configsService.getUcis(this.ucisList).subscribe(res => {
            switch (res.status) {
                case 200:
                    // console.log(res.body);
                    //store for backup
                    this.backupSettings(res.body);
                    //build device form group
                    this.buildFormGroup(this.settingsForm, res.body);

                    // console.log(this.settingsForm.value);

                    this.audioFilesService.getPregongData().subscribe(res => {
                        switch (res.status) {
                            case 200:
                                this.pregong_filename = JSON.parse(res.body)["filename"] === "" ? "N/A" : JSON.parse(res.body)["filename"];

                                this.rebooting = false;
                                this.rebooting_staticIPAddress = false;
                                this.web_proto_changed = false;
                                this.show_loading_component = false;
                                this.show_settings_component = true;
                                this.dataIsAvailable = true;

                                break;

                            case 401:
                                this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                                break;

                            case 500:
                                this.httpResponseHandler.handle500HttpErrorResponse();
                                break;

                            default:
                                console.error("Unexpected return code: ", res.status);
                        }
                    });

                    // this.rebooting = false;
                    // this.rebooting_staticIPAddress = false;
                    // this.web_proto_changed = false;
                    // this.show_loading_component = false;
                    // this.show_settings_component = true;
                    // this.dataIsAvailable = true;

                    break;

                case 401:
                    this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                    break;

                case 500:
                    this.httpResponseHandler.handle500HttpErrorResponse();
                    break;

                default:
                    console.error("Unexpected return code: ", res.status);
            }
        });
    }


    backupSettings(configDict) {
        Object.keys(configDict).forEach(key => {
            this.backupSettingsDict[key] = configDict[key];
        });
    }


    buildFormGroup(formGroup: FormGroup, uciData) {
        Object.keys(uciData).forEach(key => {
            if (key === "icpgw.barp" || key === "icpgw.rtp" || key === "icpgw.modbus") {
                Object.keys(uciData[key]).forEach(option => {
                    const uci = key + "." + option;
                    switch (uci) {
                        case "icpgw.barp.master_id":
                            formGroup.addControl('barp_gateway_id', new FormControl<number>(uciData[key][option]));
                            break;
                        case "icpgw.barp.paging_priority":
                            formGroup.addControl('barp_paging_prio', new FormControl<number>(uciData[key][option]));
                            break;
                        case "icpgw.barp.audio_port":
                            formGroup.addControl('barp_audio_port', new FormControl<number>(uciData[key][option]));
                            break;
                        case "icpgw.barp.control_port":
                            formGroup.addControl('barp_control_port', new FormControl<number>(uciData[key][option]));
                            break;
                        case "icpgw.barp.status_port":
                            formGroup.addControl('barp_status_port', new FormControl<number>(uciData[key][option]));
                            break;
                        case "icpgw.barp.sample_rate":
                            formGroup.addControl('barp_sample_rate', new FormControl<number>(parseInt(uciData[key][option])));
                            break;
                        case "icpgw.barp.audio_multicast_group_address":
                            formGroup.addControl('barp_mcast_group', new FormControl<string>(uciData[key][option]));
                            break;
                        case "icpgw.barp.ip":
                            formGroup.addControl('barp_ip_addr', new FormControl<string>(uciData[key][option]));
                            break;
                        case "icpgw.barp.format":
                            formGroup.addControl('barp_audio_format', new FormControl<string>(uciData[key][option]));
                            break;

                        case "icpgw.rtp.port":
                            formGroup.addControl('rtp_port', new FormControl<number>(uciData[key][option]));
                            break;
                        case "icpgw.rtp.format":
                            formGroup.addControl('rtp_format', new FormControl<string>(uciData[key][option]));
                            break;
                        case "icpgw.rtp.ip":
                            formGroup.addControl('rtp_ip_addr', new FormControl<string>(uciData[key][option]));
                            break;
                        case "icpgw.rtp.enabled":
                            formGroup.addControl('rtp_stream_enable', new FormControl<boolean>(uciData[key][option] === "True", Validators.required));
                            break;

                        case "icpgw.modbus.port":
                            formGroup.addControl('modbus_port', new FormControl<string>(uciData[key][option]));
                            break;
                        case "icpgw.modbus.enabled":
                            formGroup.addControl('modbus_enabled', new FormControl<boolean>(uciData[key][option] === "True", Validators.required));
                            break;

                        default:
                    }
                });
            } else if (key === "icpgw.server.timeout"){
                formGroup.addControl('icpgw_server_timeout', new FormControl<number>(uciData[key]));
            } else if (key === "icpgw.pregong.enabled"){
                formGroup.addControl('pregong_enabled', new FormControl<boolean>(uciData[key] === "True", Validators.required));
            }
        });
    }


    getUcisFromForm() {
        let uciSettingsValues = {};
        Object.keys(this.settingsForm.value).forEach(key => {
            switch (key) {
                case "barp_gateway_id":
                    uciSettingsValues["icpgw.barp.master_id"] = this.settingsForm.get(key).value;
                    break;
                case "barp_paging_prio":
                    uciSettingsValues["icpgw.barp.paging_priority"] = this.settingsForm.get(key).value;
                    break;
                case "barp_audio_port":
                    uciSettingsValues["icpgw.barp.audio_port"] = this.settingsForm.get(key).value;
                    break;
                case "barp_control_port":
                    uciSettingsValues["icpgw.barp.control_port"] = this.settingsForm.get(key).value;
                    break;
                case "barp_status_port":
                    uciSettingsValues["icpgw.barp.status_port"] = this.settingsForm.get(key).value;
                    break;
                case "barp_sample_rate":
                    uciSettingsValues["icpgw.barp.sample_rate"] = this.settingsForm.get(key).value;
                    break;
                case "barp_mcast_group":
                    uciSettingsValues["icpgw.barp.audio_multicast_group_address"] = this.settingsForm.get(key).value;
                    break;
                case "barp_ip_addr":
                    uciSettingsValues["icpgw.barp.ip"] = this.settingsForm.get(key).value;
                    break;
                case "barp_audio_format":
                    uciSettingsValues["icpgw.barp.format"] = this.settingsForm.get(key).value;
                    break;

                // case "rtp_port":
                //     uciSettingsValues["icpgw.rtp.port"] = this.settingsForm.get(key).value;
                //     break;
                // case "rtp_format":
                //     uciSettingsValues["icpgw.rtp.format"] = this.settingsForm.get(key).value;
                //     break;
                // case "rtp_ip_addr":
                //     uciSettingsValues["icpgw.rtp.ip"] = this.settingsForm.get(key).value;
                //     break;
                // case "rtp_stream_enable":
                //     uciSettingsValues["icpgw.rtp.enabled"] = this.settingsForm.get(key).value ? "True" : "False";
                //     break;

                case "icpgw_server_timeout":
                    uciSettingsValues["icpgw.server.timeout"] = this.settingsForm.get(key).value;
                    break;

                case "modbus_port":
                    uciSettingsValues["icpgw.modbus.port"] = this.settingsForm.get(key).value;
                    break;
                case "modbus_enabled":
                    uciSettingsValues["icpgw.modbus.enabled"] = this.settingsForm.get(key).value ? "True" : "False";
                    break;

                case "pregong_enabled":
                    uciSettingsValues["icpgw.pregong.enabled"] = this.settingsForm.get(key).value ? "True" : "False";
                    break;

                default:
            }
        });
        return uciSettingsValues;
    }


    cancelChanges() {
        this.show_settings_component = false;
        this.loadDataFromDevice();
    }


    submit() {
        if (this.settingsForm.valid) {
            this.show_settings_component = false;
            this.show_loading_component = true; // start loading spinner
            setTimeout(() => { //wait a bit to be able to see spinner for usability
                //build dict with ucis to send
                const payload = {
                    commit: true,
                    ucis: this.getUcisFromForm()
                }

                this.configsService.setSettings(payload).subscribe(data => {
                    switch (data.status) {
                        case 200: //SUCCESS
                            this.toast.showSuccessMessage("New configuration applied successfully.");
                            this.loadDataFromDevice();
                            break;

                        case 400: //BAD_REQUEST
                            this.toast.showErrorMessage(data.error);
                            this.loadDataFromDevice();
                            break;

                        case 401:
                            this.httpResponseHandler.handle401HttpErrorResponse((data as HttpErrorResponse).error);
                            break;

                        case 500: //INTERNAL_SERVER_ERROR
                            this.httpResponseHandler.handle500HttpErrorResponse();
                            break;

                        default:
                            console.error("Unexpected return code: ", data.status);
                    }
                });
            }, 1300);
        }
    }


    reloadWebUI() {
        this.systemFuncsService.appIsRunning().subscribe(res => {
            //console.log(res);
            if (res.status === 200) {
                window.location.href = this.settingsForm.get('web_proto').value + '://' + window.location.hostname;
            } else {
                setTimeout(() => this.reloadWebUI(), 1000);
            }
        });
    }


    scrollToId(el: HTMLElement) {
        el.scrollIntoView({'behavior': 'smooth'});
    }


    getLoadingText(): string | string[] {
        return "processing...";
    }


    openUploadFileModal() {
        const dialogConfig = new MatDialogConfig();
        // The user can't close the dialog by clicking outside its body
        dialogConfig.disableClose = false;
        dialogConfig.id = "upload-files-modal-component";
        dialogConfig.height = "200px";
        dialogConfig.width = "900px";
        dialogConfig.position = {left: '6%', top: '10%'}
        dialogConfig.panelClass = "no-padding-modal";
        dialogConfig.data = {
            "component-title": "Upload pre-gong file",
            "button-text": "Upload audio file",
            "button-msg": "Upload audio file to pre-gong",
            "toastMsgPositionClass": "",
            "MAX_FILE_SIZE": 16384000,
            "allowMultiple": false,
            // "AVAILABLE_STORAGE_SPACE": this.availableStorage,
            "AVAILABLE_STORAGE_SPACE": 17384000,
        }
        //Upload and install certificate
        this.matDialog.open(UploadFilesModalComponent, dialogConfig).afterClosed().subscribe((files) => {
            if (files) {
                this.show_settings_component = false;
                this.show_loading_component = true; // start loading spinner

                this.audioFilesService.uploadPregong(files[0]).subscribe(res => {
                    console.log(res.status);
                    switch (res.status) {
                        case 200:
                            this.toast.showSuccessMessage("New pregong file uploaded successfully.");
                            this.loadDataFromDevice();
                            break;
                        case 400:
                            this.toast.showErrorMessage(res.error);
                            this.loadDataFromDevice();
                            break;
                        case 401:
                            this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                            break;
                        case 413:
                            // this.eventFinished.emit({
                            //     "toast-msg": "Error! Maximum upload size exceeded!",
                            //     "toast-class": 'my-toast-error'
                            // });
                            break;
                        case 500:
                            this.httpResponseHandler.handle500HttpErrorResponse();
                            break;
                        case 507:
                            // this.eventFinished.emit({
                            //     "toast-msg": "Device storage is full! Please remove some files to free up some space.",
                            //     "toast-class": 'my-toast-error'
                            // });
                            break;
                        default:
                            console.error("Unexpected return code: ", res.status);
                    }
                });
            }
        });
    }


    deletePregongFile() {
        this.audioFilesService.deletePregong().subscribe(res => {
            switch (res.status) {
                case 200:
                    this.toast.showSuccessMessage("Pre-gong file removed successfully!");
                    this.loadDataFromDevice();
                    break;
                default:
                    this.toast.showErrorMessage("Error removing pre-gong file.");
                    this.loadDataFromDevice();
            }
        })
    }
}
