import {Component, EventEmitter, HostListener, OnDestroy, OnInit, Output} from '@angular/core';
import {Resampler} from "../../Resampler-Utils";
import {PagingService} from "../services/paging.service";
import {DomSanitizer} from '@angular/platform-browser';
import {ActivatedRoute, Router} from "@angular/router";
import {v4} from "uuid";
import {Utils} from "./main-bottom-utils";
import {GroupsService} from "../services/groups.service";
import {lastValueFrom} from "rxjs";

const uuid4 = v4;

@Component({
    selector: 'app-main-bottom',
    templateUrl: './main-bottom.component.html',
    styleUrls: ['./main-bottom.component.css']
})
export class MainBottomComponent implements OnInit, OnDestroy {


    @HostListener("window:beforeunload", ["$event"]) unloadHandler(event: Event) {
        if (this.paging_state === 1) {
            event.returnValue = false;
            this.stopAudioRecording();
        }
    }


    audioContext;
    audioEnabled = false;
    drawVisual;
    audioDuration;
    recordAudioStorage;
    recording = false
    recordAudio;
    duration
    playingRecordedAudio;
    playing = 0;
    paging_state = 0;
    startDate
    groupLocalStorage;
    recordingResampler;
    paging;
    musicAudio = new Audio();
    token
    uuid;
    recordingTest;
    recordingLength = 0;
    audioProcessor;
    smp_rate = -1;
    sampleRate = 8000;
    audioSource;
    analyser
    dataArrayAlt;
    resampler
    isSendingAudio = false;
    canvasCtx
    WIDTH
    HEIGHT
    bufferLengthAlt
    stream
    recordedData = []
    bufferSize = 4096
    disableStartButton = false
    disableStopButton = true
    payload;

    @Output() showRecordMode = new EventEmitter<boolean>();
    @Output() audioPlayerEnabled = new EventEmitter<boolean>();
    @Output() isSending = new EventEmitter<boolean>();
    @Output() OnClick = new EventEmitter<string>();
    @Output() OnRecordClick = new EventEmitter<string>();


    constructor(private sanitizer: DomSanitizer,
                private router: Router,
                private route: ActivatedRoute,
                private pagingService: PagingService,
                private groupsService: GroupsService) {
    }


    PARAMS = {
        "sampleRate": 24000,  // Set inside make_connection
        "bufferLength": 16 * 1024, // Value in kb, only power of 2, max 16
        "channels": 1,
        "https_audio_Server": ""
    }


    async ngOnInit() {
        this.token = this.route.snapshot.queryParamMap.get('token')
        this.audioEnabled = false;
        this.uuid = uuid4()
        this.getRecordedState()
        this.getSampleRate()
        await this.getGroups()
        this.payload = {
            "groups": this.groupLocalStorage,
            "token": this.token,
            "uuid": this.uuid,
        }
    }


    ngOnDestroy() {
        if (this.paging_state === 1)
            this.stopAudioRecording();
    }

    sanitize(url: string) {
        return this.sanitizer.bypassSecurityTrustUrl(url);
    }

    private getSampleRate() {
        this.pagingService.getSampleRate(this.token).subscribe((data) => {
            switch (data.status) {
                case 200:
                    this.PARAMS.sampleRate = data.body
                    break;
                default:
                    alert("Something Went Wrong!");
                    break;
            }
        })
    }


    private async getGroups() {
        const groupsReq = await this.groupsService.getAllGroups(this.token);
        const groupsRes = await lastValueFrom(groupsReq);
        switch (groupsRes.status) {
            case 200:
                const groupListIds = Object.keys(groupsRes.body);
                // if there is only one group, select it by default
                if (groupListIds.length === 1) {
                    this.groupLocalStorage = [groupListIds[0]];
                    localStorage.setItem('groups', JSON.stringify(this.groupLocalStorage));
                } else {
                    if (!localStorage.getItem('groups')){
                        localStorage.setItem('groups', '[]')
                        this.groupLocalStorage = [];
                    } else {
                        // Check if groups selected in local storage are all still available
                        const localStorageGroups = JSON.parse(localStorage.getItem('groups'));
                        this.groupLocalStorage = [];
                        for (const group of localStorageGroups) {
                            if (groupListIds.includes(group)){
                                this.groupLocalStorage.push(group);
                            }
                        }
                        localStorage.setItem('groups', JSON.stringify(this.groupLocalStorage));
                    }
                }
                break;
            default:
                alert("ERROR: It was not possible to get the configured groups!");
        }
    }


    private getRecordedState() {
        if (!localStorage.getItem('recordAudio')) {
            localStorage.setItem('recordAudio', 'false')
        } else {
            this.recordAudioStorage = localStorage.getItem('recordAudio') === "true";
        }
    }


    /***
     * Create all the audio context and analyser needed for the audio recording adapting to paging or recording
     * @param option paging or recording
     */
    private enablePaging(option: string) {
        return new Promise(resolve => {
            navigator.mediaDevices.getUserMedia({audio: true, video: false})
                .then((stream) => {
                    this.stream = stream;
                    this.audioContext = new AudioContext();
                    this.audioSource = this.audioContext.createMediaStreamSource(stream);
                    this.analyser = this.audioContext.createAnalyser();


                    this.analyser.minDecibels = -90;
                    this.analyser.maxDecibels = -10;
                    this.analyser.smoothingTimeConstant = 0.85;
                    this.audioSource.connect(this.analyser)


                    this.WIDTH = 300
                    this.HEIGHT = 150
                    const canvas = document.getElementById("spectrum-analyzer-canvas") as HTMLCanvasElement;
                    this.canvasCtx = canvas.getContext("2d");
                    this.canvasCtx.clearRect(0, 0, this.WIDTH, this.HEIGHT);
                    this.analyser.fftSize = 256;

                    this.bufferLengthAlt = this.analyser.frequencyBinCount;
                    this.dataArrayAlt = new Uint8Array(this.bufferLengthAlt);

                    if (option === "paging") {
                        this.audioProcessor = this.audioContext.createScriptProcessor(this.PARAMS.bufferLength, this.PARAMS.channels, this.PARAMS.channels);
                    }
                    if (option === "recording") {
                        this.audioProcessor = this.audioContext.createScriptProcessor(this.bufferSize, 1, 1);
                    }
                    this.smp_rate = this.audioContext.sampleRate
                    this.resampler = new Resampler(this.smp_rate, this.PARAMS.sampleRate, this.PARAMS.channels, this.PARAMS.bufferLength);
                    resolve(true);

                })
        })
    }


    private drawSpectrum = () => {
        if (this.recording) {
            this.drawVisual = requestAnimationFrame(this.drawSpectrum);
            this.analyser.getByteFrequencyData(this.dataArrayAlt);
            this.canvasCtx.fillStyle = "rgb(255,255,255)";

            this.canvasCtx.fillRect(0, 0, this.WIDTH, this.HEIGHT);
            const barWidth = (this.WIDTH / this.bufferLengthAlt) * 2.5;
            let barHeight;
            let x = 0;
            for (let i = 0; i < this.bufferLengthAlt; i++) {

                barHeight = this.dataArrayAlt[i];
                this.canvasCtx.fillStyle = "rgb(" + (barHeight + 100) + ",50,50)";
                this.canvasCtx.fillRect(
                    x,
                    this.HEIGHT - barHeight / 2,
                    barWidth,
                    barHeight / 2
                );
                x += barWidth + 1;
            }
        }
    };


    private processPagingAudioStream = (audio) => {
        let buffer = this.resampler.resample(audio.inputBuffer.getChannelData(0))
        let int16Arr = Utils.convertFloat32ToPcm16(buffer, this.token)
        if (this.recording && this.paging_state === 1) {
            this.pagingService.page(int16Arr.buffer, this.token).subscribe((data) => {
                switch (data.status) {
                    case 204:
                        break;
                    case 409:
                        let error_msg = "Error while paging!"
                        if (data.error){
                            error_msg += "\n" + data.error;
                        }
                        alert(error_msg);

                        this.paging_state = 0;
                        this.disableStopButton = true;
                        this.disableStartButton = false;
                        this.disconnectAudio();
                        this.paging_state = 0;
                        this.recording = false;
                        this.showRecordMode.emit(false);

                        break;
                    default:
                        alert("An error has occurred while paging!");
                        this.stopPaging();
                        break;
                }
            })
        }
    };


    private disconnectAudio() {
        this.disconnectMicrophone()
        this.audioSource.disconnect(this.audioProcessor);
        this.audioProcessor.disconnect(this.audioContext.destination);
    }


    private disconnectMicrophone(){
        this.stream.getTracks().forEach(track => {
            track.stop();
        })
        this.audioSource.disconnect(this.analyser);
    }


    private stopPaging() {
        if (this.paging_state === 1) {
            this.paging_state = 0; // Allow file playing once paging is done
            this.disableStopButton = true
            this.disableStartButton = false;
            this.disconnectAudio();
            this.pagingService.stopPaging(this.payload, this.token).subscribe((data) => {
                switch (data.status) {
                    case 204:
                        break;
                    default:
                        this.stopPaging()
                        break;
                }
            })
        }
    }


    private doPaging() {
        if (!this.playing) { // When there is no file playing
            this.pagingService.startPaging(this.payload, this.token).subscribe((data) => {
                switch (data.status) {
                    case 204:
                        this.recording = true;
                        this.showRecordMode.emit(true);
                        this.paging_state = 1;
                        this.audioSource.connect(this.audioProcessor)
                        this.audioProcessor.connect(this.audioContext.destination)
                        this.audioProcessor.onaudioprocess = this.processPagingAudioStream
                        this.drawSpectrum();
                        break;
                    default:
                    // case 409:
                        this.disableStopButton = true
                        this.disableStartButton = false;
                        this.disconnectMicrophone();
                        let error_msg = "Can not page!"
                        if (data.error){
                            error_msg += "\n" + data.error;
                        }
                        alert(error_msg);
                    //     break;
                    // default:
                    //     this.disableStopButton = true
                    //     this.disableStartButton = false;
                    //     this.disconnectMicrophone()
                    //     alert("Something went wrong!")
                    //     break;
                }
            })
        } else {
            alert("Wait for any other operations to finish and try again.")
        }

    };


    /***
     *
     * @param inputBuffer
     * @param recordingLength
     * @param my_token
     */


    private attachID(inputBuffer, recordingLength, my_token) {

        const PCM32fSamples = Utils.mergeBuffers(inputBuffer, recordingLength);
        this.recordingResampler = new Resampler(this.smp_rate, this.PARAMS.sampleRate, this.PARAMS.channels, recordingLength);
        let buffer = this.recordingResampler.resample(PCM32fSamples)
        let l = buffer.length + my_token.length;  // Add the token in the beginning
        const PCM16iSamples = new Int16Array(l);  // This buffer holds the PCM 16 audio data including the token
        for (let idx = 0; idx < my_token.length; idx++) {
            PCM16iSamples[idx] = my_token.charCodeAt(idx)
        }
        for (let i = my_token.length; i < l; i++) {
            let val = Math.floor(32767 * buffer[i - my_token.length]);
            val = Math.min(32767, val);
            val = Math.max(-32768, val);
            PCM16iSamples[i] = val
        }
        return PCM16iSamples;
    }


    async startAudioRecording() {

        if (this.recordAudioStorage) {
            this.recording = true;
            const paged = await this.enablePaging("recording");
            if (paged) {

                this.recordedData = [];
                this.showRecordMode.emit(true);
                this.doRecording()
            }
        } else {
            this.disableStopButton = false
            this.disableStartButton = true;
            const paged = await this.enablePaging("paging");
            if (paged) {
                this.doPaging();
            }
        }
    }


    stopAudioRecording() {
        this.recording = false;
        this.showRecordMode.emit(false);
        if (this.recordAudioStorage) {
            this.onRecordingStop()
            this.disconnectAudio()
            this.audioPlayerEnabled.emit(true);
        } else {
            this.stopPaging();
        }
    }


    sendAudio() {
        const audioFile = this.attachID(this.recordedData, this.recordingLength, this.token)
        this.audioEnabled = true;
        this.isSendingAudio = true;

        // let bodyStart = {"type": 3, "groups": this.groupLocalStorage, "token": this.token, "uuid": this.uuid}
        this.pagingService.startPagingRecorded(this.payload, this.token).subscribe((data) => {
            switch (data.status) {
                case 204:
                    this.pagingService.pageRecorded(audioFile.buffer, this.token).subscribe((putData) => {
                        switch (putData.status) {
                            case 204:
                                this.pagingService.stopPagingRecorded(this.payload, this.token).subscribe((endData) => {
                                    switch (endData.status) {
                                        case 204:
                                            this.isSendingAudio = false;
                                            break;
                                        default:
                                            let error_msg = "Can not stop paging recorded audio!"
                                            if (data.error){
                                                error_msg += "\n" + data.error;
                                            }
                                            alert(error_msg);
                                            break;
                                    }
                                })
                                break;
                            default: // 409
                                let error_msg = "Can not page recorded audio!"
                                if (data.error){
                                    error_msg += "\n" + data.error;
                                }
                                alert(error_msg);
                                this.isSendingAudio = false;
                        }
                    })
                    break;
                default:
                    let error_msg = "Can not page recorded audio!"
                    if (data.error){
                        error_msg += "\n" + data.error;
                    }
                    alert(error_msg);
                    this.isSendingAudio = false;
            }
        });
    }


    startPlayingAudio() {
        try {
            this.musicAudio.play().then();
            this.playingRecordedAudio = true;
        } catch (err) {
            console.log("Failed to play, error: " + err);
        }
    };


    pauseRecordedAudio() {
        try {
            this.musicAudio.pause();
            this.playingRecordedAudio = false;
        } catch (err) {
            console.log("Failed to stop, error: " + err);
        }
    };


    clear() {
        this.audioEnabled = false;
        this.recordingLength = 0;
        this.recordingTest = ""
        this.audioPlayerEnabled.emit(this.audioEnabled);
        this.showRecordMode.emit(false);
    }


    private doRecording() {
        this.audioSource.connect(this.audioProcessor)
        this.audioProcessor.connect(this.audioContext.destination)
        this.startDate = new Date();
        this.audioProcessor.onaudioprocess = this.processRecordingAudioStream
        this.drawSpectrum();
    };


    private processRecordingAudioStream = (audio) => {
        if (!this.recording) {
            return;
        }
        if (this.recording) {
            this.recordedData.push(new Float32Array(audio.inputBuffer.getChannelData(0)));
            this.recordingLength = this.recordingLength + this.bufferSize;
            this.duration = new Date().getTime() - this.startDate.getTime();
        }
    };


    private onRecordingStop() {
        let data: any = Utils.exportWav(this.recordedData,this.recordingLength,this.smp_rate)
        let blob = new Blob([data.view], {type: 'audio/wav'});
        this.recordingTest = URL.createObjectURL(blob);
        this.recordingTest = this.sanitize(this.recordingTest)
        this.musicAudio = new Audio();
        this.musicAudio.src = this.recordingTest.changingThisBreaksApplicationSecurity
        this.musicAudio.onloadedmetadata = () => {
            this.audioDuration = Utils.formatTime(this.musicAudio.duration.toFixed(0))
        }
        this.musicAudio.onended = () => {
            this.playingRecordedAudio = false;
        }
        this.audioEnabled = true;

    }


}
