import {Component, ElementRef, OnInit, ViewChild} from '@angular/core';
import {FormControl, FormGroup, UntypedFormControl, UntypedFormGroup, Validators} from "@angular/forms";
import {DeviceConfigService} from "../services/device-config.service";
import {MatSliderChange} from '@angular/material/slider';
import {HttpErrorResponse} from '@angular/common/http';
import {HTTPResponseHandlerService} from '../services/httpresponse-handler.service';
import {SystemFunctionsService} from '../services/system-functions.service';
import {SystemInfoService} from '../services/system-info.service';
import {TopBarService} from '../services/top-bar.service';
import {buildSettingsFormGroup, settingsPatternValidators} from "./settings-utils";
import {ToastMessageService} from "../services/toast-message.service";
import {ToastrService} from "ngx-toastr";
import {MatDialog, MatDialogConfig} from "@angular/material/dialog";
import {FormCanDeactivate} from "../can-deactivate/form-can-deactivate";
import DeviceTypes from '../utils/DeviceTypes';


@Component({
    selector: 'app-settings',
    templateUrl: './settings.component.html',
    styleUrls: ['../css/general.css', '../css/mat-form-field.css', './settings.component.css']
})


export class SettingsComponent extends FormCanDeactivate implements OnInit {

    /* Forms */
    settingsForm = new FormGroup<any>({});
    /* Backup */ //used to check what changed exactly, to decide which message should show while submitting
    backupSettingsDict = {};

    ucisList = {
        'alias': 'device.device.alias',
        'volume': 'application.audio.volume',
        'muted': 'application.audio.muted',
        'sonic_ip': 'sonic_ip.general.enabled',
        'sonic_ip_vol': 'sonic_ip.general.volume',
        'method': 'network.eth0.proto',
        'dhcpname': 'network.eth0.dhcpname',
        'static_ip_addr': 'network.eth0.ipaddr',
        'static_netmask': 'network.eth0.netmask',
        'static_gateway': 'network.eth0.gateway',
        'dns_type': 'network.eth0.dns_type',
        'dns_server1': 'network.eth0.dns1',
        'dns_server2': 'network.eth0.dns2',
        'timezone': 'timezone.config.value',
        'ntp_server1': 'ntp.source.server1',
        'ntp_server2': 'ntp.source.server2',
        'ntp_server3': 'ntp.source.server3',
        'defaults_opt': 'security.defaults.enabled',
        'update_opt': 'security.update.enabled',
        'reboot_opt': 'security.reset.enabled',
        'webui_pwd_flag': 'httpd.webserver.password_set',
        'ssh': 'dropbear.RunCtl.enable',
        'session_timeout': 'httpd.webserver.session_timeout',
    };

    mainVolume: FormControl<number>;
    muted: boolean;

    //WebUI Password
    webui_pwd_flag: boolean;
    pwd_set = '';
    pwd_old = '';
    pwd_chk = '';

    //API Key
    currentApiKey = '';
    api_key_timestamp: Date;
    api_key_visibility: boolean;

    //System type
    hwTypeId = null;
    hasSNMPSettings = false;
    hasPoESettings = false;

    dataIsAvailable: boolean = false;
    show_loading_component: boolean = false;
    show_settings_component: boolean = false;
    changing_volume = false;

    /* State Variables */

    // submitting status
    rebooting: boolean = false;
    rebooting_staticIPAddress: boolean = false;
    web_proto_changed: boolean = false;

    deviceLicenses = {};

    DEFAULT_SESSION_TIMEOUT_VALUE = 30; //min


    constructor(private configsService: DeviceConfigService,
                private toastr: ToastrService,
                private toast: ToastMessageService,
                private httpResponseHandler: HTTPResponseHandlerService,
                private systemInfoService: SystemInfoService,
                private topBarService: TopBarService,
                private systemFuncsService: SystemFunctionsService,
                public matDialog: MatDialog,
                private elem: ElementRef
    ) {
        super();
    }


    canDeactivateComponent() {
        return !this.settingsForm.dirty && this.pwd_old === "" && this.pwd_set === "" && this.pwd_chk === "";
    };


    ngOnInit() {
        this.toast.clearToastMessages();
        this.topBarService.updateDeviceLocalTime();
        this.loadDataFromDevice();
    }


    async loadDataFromDevice() {
        this.dataIsAvailable = false;
        this.settingsForm = new FormGroup({});
        this.configsService.getUcis(this.ucisList).subscribe(res => {
            switch (res.status) {
                case 200:
                    //store for backup
                    this.backupSettings(res.body);
                    //build device form group
                    buildSettingsFormGroup(this.settingsForm, res.body);

                    this.mainVolume = new FormControl<number>(res.body["application.audio.volume"], [Validators.min(0), Validators.max(100), Validators.required]);

                    try {
                        this.muted = Boolean(JSON.parse("" + res.body["application.audio.muted"]));
                    } catch (e) {
                        console.error("application.audio.muted not properly defined.");
                        this.muted = false;
                    }

                    this.webui_pwd_flag = res.body["httpd.webserver.password_set"] === 'true';
                    this.pwd_set = '';
                    this.pwd_old = '';
                    this.pwd_chk = '';

                    this.rebooting = false;
                    this.rebooting_staticIPAddress = false;
                    this.web_proto_changed = false;
                    this.show_loading_component = false;
                    this.show_settings_component = true;
                    this.dataIsAvailable = true;

                    if (this.webui_pwd_flag){
                        console.log("monitoring timeout session");
                        this.startMonitoringSessionTimeout();
                    } else {
                        console.log("no login: don't monitor session timeout");
                    }

                    break;

                case 401:
                    this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                    break;

                case 500:
                    this.httpResponseHandler.handle500HttpErrorResponse();
                    break;

                default:
                    console.error("Unexpected return code: ", res.status);
            }
        });
    }


    backupSettings(configDict) {
        Object.keys(configDict).forEach(key => {
            this.backupSettingsDict[key] = configDict[key];
        });
    }


    //Make value update while slider is moving
    onVolumeChange(event: MatSliderChange, formName: string) {
        if (formName === "mainVolume") {
            this.mainVolume.setValue(event.value);
        } else {
            this.settingsForm.get(formName).setValue(event.value);
        }
    }


    submitVolume() {
        if (this.changing_volume) {
            //to keep the ability to change the volume without submitting
            let uciDict = {[this.ucisList['volume']]: "" + this.mainVolume.value};
            this.configsService.setVolume(uciDict).subscribe(data => {
                switch (data.status) {
                    case 200: //SUCCESS
                        break;

                    case 400: //BAD_REQUEST
                        this.toast.showErrorMessage(data.error['msg']);
                        break;

                    case 401:
                        this.httpResponseHandler.handle401HttpErrorResponse((data as HttpErrorResponse).error);
                        break;

                    case 500: //INTERNAL_SERVER_ERROR
                        this.httpResponseHandler.handle500HttpErrorResponse();
                        break;

                    default:
                        console.error("Unexpected return code: ", data.status);
                }
                this.changing_volume = false;
            });
        }

    }


    onMuteButtonHandle() {
        // console.log("### onMuteButtonHandle", this.muted, typeof this.muted, "->", !this.muted);
        this.muted = !this.muted;
        let uciDict = {[this.ucisList['muted']]: this.muted};
        this.configsService.setMuted(uciDict).subscribe(data => {
            //console.log(data.status);
            switch (data.status) {
                case 200: //SUCCESS
                    break;

                case 400: //BAD_REQUEST
                    this.toast.showErrorMessage(data.error['msg']);
                    break;

                case 401:
                    this.httpResponseHandler.handle401HttpErrorResponse((data as HttpErrorResponse).error);
                    break;

                case 500: //INTERNAL_SERVER_ERROR
                    this.httpResponseHandler.handle500HttpErrorResponse();
                    break;

                default:
                    console.error("Unexpected return code: ", data.status);
            }
        });
    }


    cancelChanges() {
        this.show_settings_component = false;
        this.loadDataFromDevice();
    }


    submit() {
        // check if passwords are valid
        if (this.passwordsAreValid()) {
            if (this.formIsValid()) {
                this.show_settings_component = false;
                this.show_loading_component = true; // start loading spinner
                setTimeout(() => { //wait a bit to be able to see spinner for usability

                    //build dict with ucis to send
                    let settingsToSubmit = {};
                    Object.keys(this.settingsForm.value).forEach(key => {
                        if ((key === "dns_server1" || key === 'dns_server2') && this.settingsForm.get('dns_type').value === "AUTO") {
                            //don't add
                        } else if ((key === "static_ip_addr" || key === 'static_netmask' || key === 'static_gateway') && this.settingsForm.get('method').value === "dhcp") {
                            //don't add
                        } else if (key === "dhcpname" && this.settingsForm.get('method').value === "static") {
                            //don't add
                        } else if (key === "ssh") {
                            settingsToSubmit[this.ucisList[key]] = this.settingsForm.get('ssh').value ? "1" : "0";
                        } else if (key !== "date" && key !== "hours" && key !== "minutes" && key !== "seconds") {
                            settingsToSubmit[this.ucisList[key]] = "" + this.settingsForm.get(key).value;
                        }
                    });

                    //send passwords as well
                    let passwords = {}
                    if (this.pwd_old !== '') {
                        passwords['pwd_old'] = this.pwd_old;
                    }
                    if (this.pwd_set !== '') {
                        passwords['pwd_set'] = this.pwd_set;
                    }

                    const payload = {
                        commit: true
                    }

                    if (Object.keys(settingsToSubmit).length !== 0) {
                        payload['ucis'] = settingsToSubmit;
                    }

                    if (Object.keys(passwords).length !== 0) {
                        payload['pwdSettings'] = passwords;
                    }

                    // reset edited
                    this.settingsForm.markAsPristine();
                    this.pwd_old = '';
                    this.pwd_set = '';
                    this.pwd_chk = '';

                    this.configsService.setSettings(payload).subscribe(data => {
                        switch (data.status) {
                            case 200: //SUCCESS
                                this.finishSubmit();
                                // this.reloadWebUI();
                                break;

                            case 400: //BAD_REQUEST
                                this.toast.showErrorMessage(data.error);
                                this.loadDataFromDevice();
                                break;

                            case 401:
                                this.httpResponseHandler.handle401HttpErrorResponse((data as HttpErrorResponse).error);
                                break;

                            case 500: //INTERNAL_SERVER_ERROR
                                this.httpResponseHandler.handle500HttpErrorResponse();
                                break;

                            default:
                                console.error("Unexpected return code: ", data.status);
                        }
                    });
                }, 1300);
            }
        }
    }


    formIsValid() {
        if (this.settingsForm.invalid) {
            //let invalid = true;
            for (let key in this.settingsForm.controls) {
                if (this.settingsForm.get(key).invalid) {
                    if ((key === "dns_server1" || key === 'dns_server2') && this.settingsForm.get('dns_type').value === "AUTO") {
                        //invalid = false;
                    } else if ((key === "static_ip_addr" || key === 'static_netmask' || key === 'static_gateway') && this.settingsForm.get('method').value === "dhcp") {
                        //invalid = false;
                    } else {
                        this.toast.showErrorMessage("Configuration is invalid!");
                        return false;
                    }
                }
            }
            /*if (invalid) {
              this.toastr.show("Configuration is invalid!", '', {closeButton: true, disableTimeOut: true, toastClass: 'my-toast-error'});
              return false;
            }*/
        }
        if (this.settingsForm.get('ntp_server1').value === "" && this.settingsForm.get('ntp_server2').value === "" && this.settingsForm.get('ntp_server3').value === "") {
            this.toast.showErrorMessage("At least 1 NTP server should be configured!");
            return false;
        }
        if (this.settingsForm.get('dns_type').value === "MANUAL" && this.settingsForm.get('dns_server1').value === "" && this.settingsForm.get('dns_server2').value === "") {
            this.toast.showErrorMessage("At least 1 DNS server should be configured!");
            return false;
        }

        return true;
    }


    passwordsAreValid() {
        if (!this.checkPasswordsMatch()) {
            this.toast.showErrorMessage("Passwords don't match! Please correct them before submitting your changes!");
            return false;
        }
        if (this.webui_pwd_flag && this.pwd_old === '' && this.pwd_set !== '') {
            this.toast.showErrorMessage("You can not set a new password without inputting the current one!");
            return false;
        }
        return true;
    }


    checkPasswordsMatch() {
        return this.pwd_set === this.pwd_chk;
    }


    /**
     * Analyse form value changes and flag the affected services.
     */
    finishSubmit() {

        // if protocol changed to dhcp
        if (this.settingsForm.get('method').value !== this.backupSettingsDict['network.eth0.proto'] && this.settingsForm.get('method').value === "dhcp") {                                                            // if it changed to dhcp
            //console.log("network changed to dhcp");
            this.rebooting = true;
        }
        // if protocol changed to static
        if (this.settingsForm.get('method').value !== this.backupSettingsDict['network.eth0.proto'] && this.settingsForm.get('method').value === "static") {                                                   // if it changed to static
            //console.log("network changed to static");
            this.rebooting_staticIPAddress = true;
        }
        // if static ip address changed
        if (this.settingsForm.get('static_ip_addr').value !== this.backupSettingsDict['network.eth0.ipaddr'] && this.settingsForm.get('method').value === "static") {
            //console.log("static ip address changed");
            this.rebooting_staticIPAddress = true;
        }
        // if any other network setting changed
        if (this.settingsForm.get('dhcpname').value !== this.backupSettingsDict['network.eth0.dhcpname'] || this.settingsForm.get('static_netmask').value !== this.backupSettingsDict['network.eth0.netmask'] || this.settingsForm.get('static_gateway').value !== this.backupSettingsDict['network.eth0.gateway'] || this.settingsForm.get('dns_type').value !== this.backupSettingsDict['network.eth0.dns_type'] || this.settingsForm.get('dns_server1').value !== this.backupSettingsDict['network.eth0.dns1'] || this.settingsForm.get('dns_server2').value !== this.backupSettingsDict['network.eth0.dns2']) {
            //console.log("network setting changed");
            this.rebooting = true;
        }

        if (this.rebooting) {
            if (!this.web_proto_changed && !this.rebooting_staticIPAddress) {
                setTimeout(() => this.reloadWebUI(), 10000);
            } else {
                let newURL = "https://";
                if (this.rebooting_staticIPAddress) {
                    newURL += this.settingsForm.get('static_ip_addr').value;
                } else {
                    newURL += window.location.hostname;
                }
                //console.log("redirect 55000", newURL);
                setTimeout(() => window.location.href = newURL, 55000);
            }
        } else if (this.rebooting_staticIPAddress) {
            let newURL = "https://" + this.settingsForm.get('static_ip_addr').value;
            //console.log("redirect 55000", newURL);
            setTimeout(() => window.location.href = newURL, 55000);
        } else {
            //if it won't reboot neither restart backend, it can show toast message and reload the data
            this.toast.showSuccessMessage("New configuration applied successfully.");
            this.loadDataFromDevice();
        }
    }


    reloadWebUI() {
        this.systemFuncsService.appIsRunning().subscribe(res => {
            //console.log(res);
            if (res.status === 200) {
                window.location.href = 'https://' + window.location.hostname;
            } else {
                setTimeout(() => this.reloadWebUI(), 1000);
            }
        });
    }


    /**
     * Show loading component after installing an additional certificate
     */
    setShowLoading($event) {
        this.show_loading_component = $event;
        this.show_settings_component = !$event;
    }


    /**
     * Reload data and show feedback in toastr message after installing additional certificates
     */
    finishLoading(data) {
        //console.log("finish loading event");
        //console.log(data);
        this.loadDataFromDevice();
        if (data["toast-class"] === 'my-toast-warning') {
            this.toast.showWarningMessage(data["toast-msg"]);
        } else if (data["toast-class"] === 'my-toast-error') {
            this.toast.showErrorMessage(data["toast-msg"]);
        } else if (data["toast-class"] === 'my-toast-success') {
            this.toast.showSuccessMessage(data["toast-msg"]);
        }
    }


    scrollToId(el: HTMLElement) {
        el.scrollIntoView({'behavior': 'smooth'});
    }


    DHCPSettingChanged() {
        let dhcp_value = this.settingsForm.get('method').value;
        if (dhcp_value === 'static') {
            this.settingsForm.get('dns_type').setValue('MANUAL');

        }
    }


    getLoadingText(): string | string[] {
        if (!this.rebooting && !this.rebooting_staticIPAddress && !this.web_proto_changed) {
            return "processing...";
        }
        if (this.rebooting && !this.rebooting_staticIPAddress) {
            return [
                "Please wait while the device reboots. The page will be automatically reloaded.",
                "Note: If the IP Address of the device changes after reboot, please insert the new URL into the browser.",
            ];
        }
        if (this.rebooting_staticIPAddress) {
            return [
                "Please wait while the device reboots.",
                "The page will be automatically redirected to the configured static IP Address.",
            ];
        }
        return "";
    }


    timeSettingsPanelIsOpen(): boolean   {
        return window.location.hash === '#TimeSettings';
    }


    varHasValue(varName: string, varValue: string ): boolean {
        if (this.settingsForm.get(varName)) {
            return this.settingsForm.get(varName).value === varValue;
        }
        return false;
    }


    startMonitoringSessionTimeout() {
        //save session timeout in local storage
        if (this.backupSettingsDict['httpd.webserver.session_timeout'] === ''){
            //console.log("sessionTimeout", this.DEFAULT_SESSION_TIMEOUT_VALUE*60);
            localStorage.setItem('sessionTimeout', String(this.DEFAULT_SESSION_TIMEOUT_VALUE*60));
        } else {
            //console.log("sessionTimeout", parseInt(this.backupSettingsDict['httpd.webserver.session_timeout'])*60);
            localStorage.setItem('sessionTimeout', String(parseInt(this.backupSettingsDict['httpd.webserver.session_timeout'])*60));
        }

        const sessionTimeoutInterval = setInterval(() => {
            //console.log("check session timeout");
            if ( Math.floor(Date.now() / 1000) - parseInt(localStorage.getItem('lastActivity')) > parseInt(localStorage.getItem('sessionTimeout')) ){
                // has expired
                //console.log("expired!");
                this.httpResponseHandler.handle401HttpErrorResponse("Session expired");
                clearInterval(sessionTimeoutInterval);
            }
        }, 1000);

    }

}
