import {Component, OnInit, ViewChild} from '@angular/core';
import {UsersManagementService} from "../services/users-management.service";
import {HttpErrorResponse} from "@angular/common/http";
import {HTTPResponseHandlerService} from "../services/httpresponse-handler.service";
import {MatPaginator} from "@angular/material/paginator";
import {MatTableDataSource} from "@angular/material/table";
import {lastValueFrom} from "rxjs";
import {MatDialog} from "@angular/material/dialog";
import {NewOrEditUserComponent} from "../new-or-edit-user/new-or-edit-user.component";
import {ToastMessageService} from "../services/toast-message.service";
import {UserDetailsModalComponent} from "../user-details-modal/user-details-modal.component";
import {Sort} from "@angular/material/sort";
import {compareStrings} from "../utils/comparisons";
import {AudioFilesService} from "../services/audio-files.service";
import {AudioFile} from "../audio-files-tab/audio-files-tab.component";

@Component({
  selector: 'app-users-management-tab',
  templateUrl: './users-management-tab.component.html',
  styleUrls: ['../css/general.css', './users-management-tab.component.css']
})
export class UsersManagementTabComponent implements OnInit {

    MAX_N_USERS = 50;

    users : [];

    all_checked : boolean = false;
    usersSelected = []; // list of ids

    dataIsAvailable: boolean = false;

    /*** Table variables ***/
    displayedColumns = ["selection", "name", "email", "access_url", "action_buttons"];
    dataSource: MatTableDataSource<User>;

    /*** Paginator variables ***/
    @ViewChild('paginator') paginator: MatPaginator;
    pageIndex : number = 0;
    length : number;
    pageSize : number = 10;
    pageSizeOptions = [5, 10, 25];
    showFirstLastButtons : boolean = true;

    /*** Audio Files ***/
    audioFiles : AudioFile[];

    constructor(private usersService: UsersManagementService,
              private httpResponseHandler: HTTPResponseHandlerService,
                public dialog: MatDialog,
                private toast: ToastMessageService,
                private audioFilesService: AudioFilesService) { }

  ngOnInit() {
    this.loadData();
  }

  async loadData() {
      await this.getUsers();
      this.length = this.dataSource.data.length;
      this.usersSelected = [];
      this.all_checked = false;
      this.dataIsAvailable = true;
      setTimeout(() => { //small trick to have the paginator HTML element rendered before assigning it to the table
          this.dataSource.paginator = this.paginator;
      }, 100);
      await this.getAudioFiles();
  }

  async getUsers() {
    const usersReq = this.usersService.getUsers();
    const usersRes = await lastValueFrom(usersReq);
    switch (usersRes.status) {
        case 200:
            this.users = JSON.parse(usersRes.body)["users"];
            this.dataSource = new MatTableDataSource<User>(this.users);
            break;

        case 401:
            this.httpResponseHandler.handle401HttpErrorResponse((usersRes as HttpErrorResponse).error);
            break;

        case 500:
            this.httpResponseHandler.handle500HttpErrorResponse();
            break;

        default:
            console.error("Unexpected return code: ", usersRes.status);
    }
  }

    async getAudioFiles() {
        let audioFiles = [];
        const groupsReq = await this.audioFilesService.getAudioFiles();
        const groupsRes = await lastValueFrom(groupsReq);
        switch (groupsRes.status) {
            case 200:
                this.audioFiles = JSON.parse(groupsRes.body)["audio-files"];
                break;
            case 401:
                this.httpResponseHandler.handle401HttpErrorResponse((groupsRes as HttpErrorResponse).error);
                break;

            case 500:
                this.httpResponseHandler.handle500HttpErrorResponse();
                break;

            default:
                console.error("Unexpected return code: ", groupsRes.status);
        }
        return audioFiles;
    }

  getAccessUrl(token) {
      return "https://"+window.location.hostname + "/paging?token="+token;
  }

  /*** Selection ***/

  userIsSelected(userId){
      return this.usersSelected.indexOf(userId) > -1;
  }

  selectOrDeselectUser(userId){
    const idx = this.usersSelected.indexOf(userId);
    if(idx > -1){
        this.usersSelected.splice(idx, 1);
    } else {
        if (userId !== '0')
            this.usersSelected.push(userId);
    }
      this.all_checked = this.usersSelected.length === this.dataSource.data.length;
  }

  selectOrDeselectAll(){
    if (this.all_checked){
        this.usersSelected = [];
        this.all_checked = false;
    } else {
        // select all
        this.dataSource.data.forEach(userElem => {
            if (!this.userIsSelected(userElem.id) && userElem.id !== '0') this.usersSelected.push(userElem.id);
        });
        this.all_checked = true;
    }
  }

  /*** End selection ***/

  openNewOrEditUserDialog(data){
      return this.dialog.open(NewOrEditUserComponent, {
          width: "45vw",
          height: "81vh",
          data: data
      });
  }

  createNewUser() {
      const data = {
          "mode": "new",
          "audio_files": this.audioFiles
      }
      const dialogRef = this.openNewOrEditUserDialog(data);

      dialogRef.afterClosed().subscribe(async result => {
          if (result){
              // send request to save new user
              const req = this.usersService.saveNewUser(result);
              const res = await lastValueFrom(req);
              switch (res.status) {
                  case 201:
                      // reload data
                      this.dataIsAvailable = false;
                      this.loadData();
                      break;
                  case 400:
                      this.toast.showErrorMessage((res as HttpErrorResponse).error);
                      break;
                  case 401:
                      this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                      break;
                  case 500:
                      this.httpResponseHandler.handle500HttpErrorResponse();
                      break;
                  default:
                      console.error("Unexpected return code: ", res.status);
              }
          } else {
              console.log("canceled");
          }
      });
  }

  async deleteSelectedUsers() {
      if (this.usersSelected.length === 0){
          alert("Please select one or more users to delete them.");
      } else {
          if (confirm("Are you sure you want to delete the selected users?") === true) {
              const req = this.usersService.deleteUsers(this.usersSelected);
              const res = await lastValueFrom(req);
              switch (res.status) {
                  case 200:
                      // reload data
                      this.dataIsAvailable = false;
                      this.loadData();
                      break;
                  case 401:
                      this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                      break;
                  case 500:
                      this.httpResponseHandler.handle500HttpErrorResponse();
                      break;
                  default:
                      console.error("Unexpected return code: ", res.status);
              }
          }
      }
  }

  editUser(user) {
      const data = {
          "mode": "edit",
          "user": user,
          "audio_files": this.audioFiles
      }
      const dialogRef = this.openNewOrEditUserDialog(data);

      dialogRef.afterClosed().subscribe(async result => {
          if (result){
              result["id"] = user.id;
              // send request to save new user
              const req = this.usersService.saveUserChanges(result);
              const res = await lastValueFrom(req);
              switch (res.status) {
                  case 200:
                      // reload data
                      this.dataIsAvailable = false;
                      this.loadData();
                      break;
                  case 400:
                      this.toast.showErrorMessage((res as HttpErrorResponse).error);
                      break;
                  case 401:
                      this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                      break;
                  case 500:
                      this.httpResponseHandler.handle500HttpErrorResponse();
                      break;
                  default:
                      console.error("Unexpected return code: ", res.status);
              }
          } else {
              console.log("canceled");
          }

      });
  }

  renewToken(user){
    if (confirm("Are you sure you want to generate a new access token to user " + user.name + "?") === true){
        this.usersService.renewToken(user.id).subscribe( res => {
            switch (res.status) {
                case 200:
                    // reload data
                    this.dataIsAvailable = false;
                    this.loadData();
                    break;
                case 400:
                    this.toast.showErrorMessage((res as HttpErrorResponse).error);
                    break;
                case 401:
                    this.httpResponseHandler.handle401HttpErrorResponse((res as HttpErrorResponse).error);
                    break;
                case 500:
                    this.httpResponseHandler.handle500HttpErrorResponse();
                    break;
                default:
                    console.error("Unexpected return code: ", res.status);
            }
        });
    }
  }

  showUserInfoModal(user){
      this.dialog.open(UserDetailsModalComponent, {
          width: "30vw",
          height: "65vh",
          data: {
              name: user.name,
              email: user.email,
              access_url: this.getAccessUrl(user.secret)
          }
      });
  }

  downloadUsersData() {
      let csvContent = "data:text/csv;charset=utf-8,";

      csvContent += "Name,Email,Access URL\r\n";

      this.dataSource.data.forEach( user => {
          csvContent += user["name"] + "," + user["email"] + "," + this.getAccessUrl(user["secret"]) + "\r\n";
      });

      const encodedUri = encodeURI(csvContent);
      const link = document.createElement("a");
      link.setAttribute("href", encodedUri);
      link.setAttribute("download", "users.csv");
      document.body.appendChild(link);

      link.click();
  }

    /*** Sort ***/
    sortData(sort: Sort) {
        const data = this.users.slice();
        if (!sort.active || sort.direction === '') {
            this.dataSource.data = data;
            return;
        }

        this.dataSource.data = data.sort((a: User, b:User) => {
            const isAsc = sort.direction === 'asc';
            switch (sort.active) {
                case 'name':
                    return compareStrings(a.name, b.name, isAsc);
                default:
                    return 0;
            }
        });
    }
}

export interface User {
    id: string;
    name: string;
    email: string;
    access_url: string;
    audio_files: string[];
}
